# SPDX-License-Identifier: BSD-2-Clause
""" Compare two specification sets. """

# Copyright (C) 2019, 2026 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import dataclasses
import itertools
import re
from typing import Any, Callable, Iterator

from specitems import (Item, ItemCache, ItemGetValueContext, ItemValueProvider,
                       ItemSelection, SphinxContent, TextContent, TextMapper,
                       get_reference, make_label)
from specware import (gather_benchmarks_and_test_suites, gather_related_items,
                      gather_test_cases, get_constraint_items,
                      get_interface_items, get_items_by_type_map,
                      get_items_by_types, get_requirement_items,
                      recursive_is_enabled, run_command)

from .directorystate import DirectoryState
from .itemcachestate import ItemCacheDirectoryState
from .sphinxbuilder import SphinxBuilder
from .pkgitems import BuildItem, PackageBuildDirector

_Commit = dict[str, Any]

_COMMIT = re.compile(r"commit ([0-9a-f]+)$")

_EMPTY_SCOPE_TO_UIDS: dict[str, frozenset[str]] = {
    "all": frozenset(),
    "interfaces": frozenset(),
    "requirements": frozenset(),
    "validations": frozenset(),
    "unit-and-integration-tests": frozenset(),
}


def _get_items(item_cache: ItemCache, selection: ItemSelection,
               enabled_set_actions: list[dict[str, Any]],
               root_uid: str) -> list[Item]:
    selection = ItemSelection(item_cache, selection.enabled_set,
                              recursive_is_enabled)
    for action in enabled_set_actions:
        selection.apply_action(action)
    with item_cache.selection(selection):
        return gather_related_items(item_cache[root_uid])


def _get_tests(item_cache: ItemCache, uids: list[str]) -> Iterator[str]:
    for uid in uids:
        test_suites: list[Item] = []
        gather_benchmarks_and_test_suites(item_cache[uid], test_suites)
        yield from (item.uid for item in test_suites)
        for test_suite in test_suites:
            test_cases: list[Item] = []
            gather_test_cases(test_suite, test_cases)
            yield from (item.uid for item in test_cases)


def _get_other_validations(
        items_by_type: dict[str, list[Item]]) -> Iterator[str]:
    yield from (item.uid
                for item in get_items_by_types(items_by_type, (
                    "validation/by-analysis", "validation/by-inspection",
                    "validation/by-review-of-design")))


@dataclasses.dataclass
class _SpecRevision:
    """ Represents a specification revision. """

    def __init__(self, state: ItemCacheDirectoryState,
                 selection: ItemSelection, data: dict,
                 previous_uids: dict[str, frozenset[str]]) -> None:
        item_cache = state.cache()
        items = _get_items(item_cache, selection, data["enabled-set-actions"],
                           data["root-uid"])
        items_by_type = get_items_by_type_map(items)
        self.previous_uids = previous_uids
        self.uids = {
            "all":
            frozenset(item.uid for item in items),
            "interfaces":
            frozenset(item.uid for item in get_interface_items(items_by_type)),
            "requirements":
            frozenset(item.uid for item in itertools.chain(
                get_constraint_items(items_by_type),
                get_requirement_items(items_by_type))),
            "validations":
            frozenset(
                itertools.chain(
                    _get_tests(item_cache, data["validation-test-suites"]),
                    _get_other_validations(items_by_type))),
            "unit-and-integration-tests":
            frozenset(
                _get_tests(item_cache,
                           data["unit-and-integration-test-suites"])),
        }
        self.key: str = data["revision-key"]
        self.label: str = data["label"]
        self.name: str = data["revision-name"]
        self.revision: str = state.input("source")["commit"]
        self.spec_paths = state.spec_paths("")


@dataclasses.dataclass
class _CompareSpecsConfig:
    """ Represents a specification comparison configuration. """
    repository: DirectoryState
    ignored_commits: list[str]
    previous: _SpecRevision
    current: _SpecRevision


def _get_git_log(config: _CompareSpecsConfig,
                 spec_paths: list[str]) -> list[str]:
    stdout: list[str] = []
    status = run_command([
        "git", "log", "--no-renames", "-p",
        f"{config.previous.revision}..{config.current.revision}", "--"
    ] + spec_paths, config.repository.directory, stdout)
    assert status == 0
    return stdout


class _LogParser:
    # pylint: disable=too-few-public-methods
    # pylint: disable=too-many-instance-attributes

    def __init__(self, uids: frozenset[str], ignored_commits: list[str],
                 spec_paths: list[str]):
        self._uids = uids
        self._ignored_commits = ignored_commits
        self._spec_paths = sorted(spec_paths, reverse=True)
        self._commits: list[_Commit] = []
        self._current_commit: dict[str, Any] = {}
        self._current_chunk: list[str] = []
        self._current_diff: dict[str, Any] = {}
        self.consume = self._expect_commit

    def _get_diff_uid(self, diff: dict, which: str) -> str | None:
        if diff[which] == "/dev/null":
            return None
        path = diff[which][2:]
        for prefix in self._spec_paths:
            uid = path.removeprefix(prefix)
            if path != uid and path.endswith(".yml"):
                return uid[:-4]
        raise ValueError

    def _finalize_diff(self) -> None:
        diff = self._current_diff
        self._current_diff = {}
        if diff:
            try:
                a_uid = self._get_diff_uid(diff, "a")
            except ValueError:
                return
            try:
                b_uid = self._get_diff_uid(diff, "b")
            except ValueError:
                return
            if b_uid is not None and b_uid not in self._uids:
                return
            diff["a-uid"] = a_uid
            diff["b-uid"] = b_uid
            self._current_commit["diffs"].append(diff)

    def _finalize(self) -> None:
        if self._current_chunk:
            self._current_diff["chunks"].append(self._current_chunk)
            self._current_chunk = []
        self._finalize_diff()
        commit = self._current_commit
        self._current_commit = {}
        if commit and commit["diffs"]:
            if commit["commit"] in self._ignored_commits:
                commit["diffs"] = []
            self._commits.append(commit)

    def finalize(self) -> list[_Commit]:
        """ Finalizes the parsing and returns the commits. """
        self._finalize()
        return self._commits

    def _expect_commit(self, line: str) -> None:
        mobj = _COMMIT.match(line)
        assert mobj
        self._finalize()
        self._current_commit = {
            "diffs": [],
            "commit": mobj.group(1),
            "message": []
        }
        self.consume = self._expect_author

    def _expect_author(self, line: str) -> None:
        key = "Author: "
        assert line.startswith(key)
        self._current_commit["author"] = line[len(key):]
        self.consume = self._expect_date

    def _expect_date(self, line: str) -> None:
        key = "Date: "
        assert line.startswith(key)
        self._current_commit["date"] = line[len(key):]
        self.consume = self._expect_empty_line

    def _expect_empty_line(self, line: str) -> None:
        assert not line
        self.consume = self._message

    def _message(self, line: str) -> None:
        space = "    "
        if line.startswith(space):
            self._current_commit["message"].append(line[len(space):])
            return
        self.consume = self._diff_start
        self._diff_start(line)

    def _diff_start(self, line: str) -> None:
        self._finalize_diff()
        if line.startswith("---"):
            self._current_diff = {"a": line[4:], "chunks": []}
            self.consume = self._expect_diff_plus

    def _expect_diff_plus(self, line: str) -> None:
        assert line.startswith("+++")
        self._current_diff["b"] = line[4:]
        self.consume = self._expect_chunk_start

    def _expect_chunk_start(self, line: str) -> None:
        assert line.startswith("@@")
        self._current_chunk = [line]
        self.consume = self._chunk

    def _chunk(self, line: str) -> None:
        if line.startswith((" ", "-", "+")):
            self._current_chunk.append(line)
            return
        self._current_diff["chunks"].append(self._current_chunk)
        self._current_chunk = []
        if line.startswith("@@"):
            self.consume = self._expect_chunk_start
            self._expect_chunk_start(line)
            return
        if line.startswith("commit"):
            self.consume = self._expect_commit
            self._expect_commit(line)
            return
        if line.startswith("diff"):
            self.consume = self._diff_start
            self._diff_start(line)
            return


def _get_commits(config: _CompareSpecsConfig) -> list[_Commit]:
    uids_current = config.current.uids["all"]
    uids_previous = config.previous.uids["all"]
    all_uids = uids_current.union(uids_previous)
    spec_paths = sorted(
        set(
            itertools.chain(config.current.spec_paths,
                            config.previous.spec_paths)))
    parser = _LogParser(all_uids, config.ignored_commits, spec_paths)
    for line in _get_git_log(config, spec_paths):
        parser.consume(line.rstrip("\n"))
    return parser.finalize()


def _add_change_log(content: TextContent, config: _CompareSpecsConfig) -> None:
    with content.label_scope(config.current.label):
        for commit in reversed(_get_commits(config)):
            with content.section(commit["message"][0]):
                content.add(commit["message"][2:])
                if not commit["diffs"]:
                    with content.directive("note"):
                        content.add("The modifications of this change are "
                                    "not listed in the document.")
                    continue
                for diff in commit["diffs"]:
                    a_uid = diff["a-uid"]
                    b_uid = diff["b-uid"]
                    if a_uid is None:
                        with content.section(f"spec:{b_uid}"):
                            content.add(
                                "This change added the specification item.")
                        continue
                    if b_uid is None:
                        with content.section(f"spec:{a_uid}"):
                            content.add(
                                "This change removed the specification item.")
                        continue
                    with content.section(f"spec:{b_uid}"):
                        line_number_start = 1
                        for chunk in diff["chunks"]:
                            content.add_code_block(
                                chunk,
                                language="diff",
                                line_number_start=line_number_start)
                            line_number_start += len(chunk)


def _ref(uid: str, present: frozenset[str]) -> str:
    if uid in present:
        return get_reference(make_label(f"Spec{uid}"), uid)
    return uid


class CompareSpecsRegistry(BuildItem):
    """ Provides a registry for specification comparisons. """

    def __init__(self, director: PackageBuildDirector, item: Item):
        super().__init__(director, item)
        self._revisions: list[_SpecRevision] = []
        self._key_to_revision: dict[str, _SpecRevision] = {}
        self._item_history: dict[str, dict] = {}

    def _initialize_item_history(self) -> None:
        repository = self.input("repository")
        assert isinstance(repository, DirectoryState)
        previous = self._revisions[0]
        for current in self._revisions[1:]:
            uid_to_commit_list: dict[str, list[dict[str, dict]]] = {}
            config = _CompareSpecsConfig(
                repository=repository,
                ignored_commits=self["ignored-commits"],
                previous=previous,
                current=current)
            for commit in reversed(_get_commits(config)):
                uid_to_commit: dict[str, dict] = {}
                for diff in commit["diffs"]:
                    a_uid = diff["a-uid"]
                    b_uid = diff["b-uid"]
                    if b_uid is None:
                        uid = a_uid
                    else:
                        uid = b_uid
                    uid_to_commit.setdefault(uid, {
                        "message": commit["message"],
                        "diffs": []
                    })["diffs"].append(diff)
                for uid, item_commit in uid_to_commit.items():
                    uid_to_commit_list.setdefault(uid, []).append(item_commit)
            for uid, commit_list in uid_to_commit_list.items():
                history = self._item_history.setdefault(
                    uid, {
                        "changes": [],
                        "revisions": []
                    })
                history["changes"].append({
                    "commits": commit_list,
                    "key": current.key,
                    "name": current.name,
                    "revision": current,
                    "previous-revision": previous
                })
                history["revisions"].append(current.key)
            previous = current

    def _initialize_revisions(self) -> None:
        if self._revisions:
            return
        selection = self.item.cache.active_selection
        previous_uids = _EMPTY_SCOPE_TO_UIDS
        for link, state in self.input_links("spec-compare-revision"):
            assert isinstance(state, ItemCacheDirectoryState)
            data = self.substitute(link.data)
            revision = _SpecRevision(state, selection, data, previous_uids)
            self._revisions.append(revision)
            self._key_to_revision[revision.key] = revision
            previous_uids = revision.uids
        self._initialize_item_history()

    def _get_revision(self, key: str) -> _SpecRevision:
        try:
            return self._key_to_revision[key]
        except KeyError as err:
            available = [revision.key for revision in self._revisions]
            raise ValueError("there is no revision associated with the "
                             f"key '{key}', available {available}") from err

    def add_change_log(self, content: TextContent, previous_key: str,
                       current_key: str) -> None:
        """
        Add a list of changes from the specification associated with the
        previous key to the specification associated with the current key.
        """
        self._initialize_revisions()
        repository = self.input("repository")
        assert isinstance(repository, DirectoryState)
        previous = self._get_revision(previous_key)
        current = self._get_revision(current_key)
        config = _CompareSpecsConfig(repository=repository,
                                     ignored_commits=self["ignored-commits"],
                                     previous=previous,
                                     current=current)
        _add_change_log(content, config)

    def add_change_log_by_args(self, content: TextContent, args: str) -> None:
        """ Add a list of changes specified by args. """
        previous_key, _, current_key = args.partition("..")
        self.add_change_log(content, previous_key, current_key)

    def add_item_changes(self, content: TextContent, uid: str) -> None:
        """ Get the item history by UID. """
        self._initialize_revisions()
        history = self._item_history.get(uid)
        if history is None:
            content.add(
                f"There are no changes since {self._revisions[0].name}.")
            return
        for change in reversed(history["changes"]):
            name = change["name"]
            in_current = uid in change["revision"].uids["all"]
            in_previous = uid in change["previous-revision"].uids["all"]
            if not in_previous and in_current:
                content.add(f"The item is new in {name}.")
                continue
            if in_previous and not in_current:
                content.add(f"The item was deleted in {name}.")
                continue
            content.add(f"The following changes are associated with {name}.")
            for commit in reversed(change["commits"]):
                message = commit["message"]
                with content.directive("topic", message[0]):
                    if len(message) >= 2:
                        content.append(message[2:])
                    else:
                        content.append("No commit message.")
                for diff in commit["diffs"]:
                    line_number_start = 1
                    for chunk in diff["chunks"]:
                        content.add_code_block(
                            chunk,
                            language="diff",
                            line_number_start=line_number_start)
                        line_number_start += len(chunk)

    def add_changes_by_scope(self, content: TextContent, scope: str,
                             revision_key: str) -> None:
        """
        Add the changes of items within the scope from the revision associated
        with the key compared to the previous revision.
        """
        # pylint: disable=too-many-locals
        self._initialize_revisions()
        current = self._get_revision(revision_key)
        current_scope = current.uids[scope]
        current_other = current.uids["all"].difference(current_scope)
        previous_scope = current.previous_uids[scope]
        previous_other = current.previous_uids["all"].difference(
            previous_scope)
        maybe_deleted = previous_scope.difference(current_scope)
        moved_out = maybe_deleted.intersection(current_other)
        deleted = maybe_deleted.difference(moved_out)
        maybe_new = current_scope.difference(previous_scope)
        moved_in = maybe_new.intersection(previous_other)
        new = maybe_new.difference(moved_in)
        modified = sorted(uid
                          for uid in current_scope.intersection(previous_scope)
                          if current.key in self._item_history.get(
                              uid, {}).get("revisions", tuple()))
        present = self._revisions[-1].uids[scope]
        rows = [("Items", "Status")]
        rows.extend((_ref(uid, present), "Deleted") for uid in sorted(deleted))
        rows.extend(
            (_ref(uid, present), "Moved out") for uid in sorted(moved_out))
        rows.extend(
            (_ref(uid, present), "Moved in") for uid in sorted(moved_in))
        rows.extend((_ref(uid, present), "New") for uid in sorted(new))
        rows.extend((_ref(uid, present), "Modified") for uid in modified)
        assert isinstance(content, SphinxContent)
        content.add_grid_table(rows, widths=[80, 20], font_size=-2)

    def add_changes_by_scope_by_args(self, content: TextContent,
                                     args: str) -> None:
        """ Add the changes specified by args. """
        scope, _, revision_key = args.partition(":")
        self.add_changes_by_scope(content, scope, revision_key)


class CompareSpecsProvider(ItemValueProvider):
    """ Provides a specification comparison. """

    # pylint: disable=too-few-public-methods

    def __init__(self, builder: SphinxBuilder) -> None:
        super().__init__(builder.mapper)
        self._builder = builder
        self.mapper.add_get_value("pkg/spec-compare-registry:/spec-change-log",
                                  self._get_spec_change_log)
        self.mapper.add_get_value(
            "pkg/spec-compare-registry:/spec-item-changes",
            self._get_spec_item_changes)
        self.mapper.add_get_value(
            "pkg/spec-compare-registry:/spec-changes-by-scope",
            self._get_spec_changes_by_scope)

    def _get_spec_things(
            self, ctx: ItemGetValueContext,
            what: Callable[[CompareSpecsRegistry, TextContent, str],
                           None]) -> str:
        builder = self._builder
        director = builder.director
        with builder.section_level_scope(ctx) as args:
            assert isinstance(self.mapper, TextMapper)
            content = self.mapper.create_content(
                section_level=builder.section_level)
            registry = director[ctx.item.uid]
            assert isinstance(registry, CompareSpecsRegistry)
            assert args
            what(registry, content, args)
            return "\n".join(content)

    def _get_spec_change_log(self, ctx: ItemGetValueContext) -> str:
        return self._get_spec_things(
            ctx, CompareSpecsRegistry.add_change_log_by_args)

    def _get_spec_item_changes(self, ctx: ItemGetValueContext) -> str:
        return self._get_spec_things(ctx,
                                     CompareSpecsRegistry.add_item_changes)

    def _get_spec_changes_by_scope(self, ctx: ItemGetValueContext) -> str:
        return self._get_spec_things(
            ctx, CompareSpecsRegistry.add_changes_by_scope_by_args)



# SPDX-License-Identifier: BSD-2-Clause
""" Provides the default build item factory. """

# Copyright (C) 2023, 2026 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from .archiver import Archiver
from .directorystate import DirectoryState, RepositoryState
from .docbuilder import DocumentBuilder
from .doxyfile import Doxyfile
from .gcdaproducer import GCDAProducer
from .icdbuilder import ICDBuilder
from .itemcachestate import ItemCacheDirectoryState
from .linkhub import LinkHub
from .membenchcollector import MembenchCollector
from .pkgitems import (BuildItemFactory, GenericPackageComponent, Redirection)
from .packagechanges import PackageChanges
from .packagemanual import PackageManualBuilder, PackageSummary
from .perfimages import BuildPerformanceImages
from .reposubset import RepositorySubset
from .rtems import RTEMSItemCache, RTEMSPackageComponent
from .rtemstestsimages import BuildRTEMSTestsImages
from .runactions import RunActions
from .runexecutablecmd import RunExecutableCommand
from .sddlinker import SDDLinker
from .spamrmanager import SpamrManager
from .speccompare import CompareSpecsRegistry
from .sreldbuilder import SRelDBuilder
from .srsbuilder import SRSBuilder
from .stdtailoring import ECSSClause
from .svrbuilder import SVRBuilder
from .testaggregator import TestAggregator
from .testplanbuilder import TestPlanBuilder
from .testreporter import TestReporter
from .testrunner import (TestRunner, GRMONManualTestRunner,
                         SubprocessTestRunner, TestLog)
from .testrunneresa import ESATestRunner
from .testrunnergrpc import GRPCTestRunner


def create_build_item_factory() -> BuildItemFactory:
    """ Create the default build item factory. """
    factory = BuildItemFactory()
    factory.add_constructor("pkg/component/generic", GenericPackageComponent)
    factory.add_constructor("pkg/component/rtems", RTEMSPackageComponent)
    factory.add_constructor("pkg/directory-state/archive", Archiver)
    factory.add_constructor("pkg/directory-state/doxyfile", Doxyfile)
    factory.add_constructor("pkg/directory-state/explicit", DirectoryState)
    factory.add_constructor("pkg/directory-state/gcda-producer", GCDAProducer)
    factory.add_constructor("pkg/directory-state/item-cache",
                            ItemCacheDirectoryState)
    factory.add_constructor("pkg/directory-state/membench-collector",
                            MembenchCollector)
    factory.add_constructor("pkg/directory-state/package-summary",
                            PackageSummary)
    factory.add_constructor("pkg/directory-state/patterns", DirectoryState)
    factory.add_constructor("pkg/directory-state/performance-images",
                            BuildPerformanceImages)
    factory.add_constructor("pkg/directory-state/repository", RepositoryState)
    factory.add_constructor("pkg/directory-state/repository-subset",
                            RepositorySubset)
    factory.add_constructor("pkg/directory-state/rtems-tests-images",
                            BuildRTEMSTestsImages)
    factory.add_constructor("pkg/directory-state/run-executable-command",
                            RunExecutableCommand)
    factory.add_constructor("pkg/directory-state/sdd-linker", SDDLinker)
    factory.add_constructor("pkg/directory-state/sphinx/ddf-sreld",
                            SRelDBuilder)
    factory.add_constructor("pkg/directory-state/sphinx/djf-svr", SVRBuilder)
    factory.add_constructor("pkg/directory-state/sphinx/generic",
                            DocumentBuilder)
    factory.add_constructor("pkg/directory-state/sphinx/package-manual",
                            PackageManualBuilder)
    factory.add_constructor("pkg/directory-state/sphinx/paf-spamr",
                            SpamrManager)
    factory.add_constructor("pkg/directory-state/sphinx/test-plan",
                            TestPlanBuilder)
    factory.add_constructor("pkg/directory-state/sphinx/test-report",
                            TestReporter)
    factory.add_constructor("pkg/directory-state/sphinx/ts-icd", ICDBuilder)
    factory.add_constructor("pkg/directory-state/sphinx/ts-srs", SRSBuilder)
    factory.add_constructor("pkg/directory-state/test-coverage",
                            DirectoryState)
    factory.add_constructor("pkg/directory-state/test-log", TestLog)
    factory.add_constructor("pkg/directory-state/unpacked-archive",
                            DirectoryState)
    factory.add_constructor("pkg/link-hub", LinkHub)
    factory.add_constructor("pkg/package-changes", PackageChanges)
    factory.add_constructor("pkg/redirection", Redirection)
    factory.add_constructor("pkg/rtems-item-cache", RTEMSItemCache)
    factory.add_constructor("pkg/run-actions", RunActions)
    factory.add_constructor("pkg/spec-compare-registry", CompareSpecsRegistry)
    factory.add_constructor("pkg/test-aggregator", TestAggregator)
    factory.add_constructor("pkg/test-runner/esa", ESATestRunner)
    factory.add_constructor("pkg/test-runner/dummy", TestRunner)
    factory.add_constructor("pkg/test-runner/grmon-manual",
                            GRMONManualTestRunner)
    factory.add_constructor("pkg/test-runner/grpc", GRPCTestRunner)
    factory.add_constructor("pkg/test-runner/subprocess", SubprocessTestRunner)
    factory.add_constructor("requirement/non-functional/ecss", ECSSClause)
    return factory



# SPDX-License-Identifier: BSD-2-Clause
""" Builds a Software Requirements Specification (SRS). """

# Copyright (C) 2022, 2026 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from specitems import Item, ItemGetValueContext, SphinxContent

from .pkgitems import PackageBuildDirector
from .specdocbuilder import SpecDocumentBuilder


def _add_no_requirements(content: SphinxContent, which: str) -> None:
    content.add(f"""At the time of pre-qualification, no specific {which}
requirements have been identified, but they shall be determined as
appropriate for each individual mission or system in which RTEMS is
integrated.""")


class SRSBuilder(SpecDocumentBuilder):
    """ Builds a Software Requirements Specification (SRS). """

    def __init__(self, director: PackageBuildDirector, item: Item) -> None:
        super().__init__(director, item)
        my_type = self.item.type
        self.mapper.add_get_value(f"{my_type}:/srs-requirements",
                                  self._get_requirements)
        self.mapper.add_get_value(f"{my_type}:/srs-constraints",
                                  self._get_constraints)
        self.mapper.add_get_value(
            "requirement/functional/action:/text-template",
            self._get_action_template)
        perf_runtime = "requirement/non-functional/performance-runtime"
        self.mapper.add_get_value(f"{perf_runtime}:/environment",
                                  self._get_environment)
        self.mapper.add_get_value(f"{perf_runtime}:/limit-kind",
                                  self._get_limit_kind)
        self.mapper.add_get_value(f"{perf_runtime}:/limit-condition",
                                  self._get_limit_condition)
        self.mapper.add_get_value(
            "runtime-measurement-test:/params/buffer-count", self._get_count)
        self.mapper.add_get_value(
            "runtime-measurement-test:/params/sample-count", self._get_count)

    def get_items_of_document(self) -> list[Item]:
        return self.spec.get_related_requirements()

    def _get_constraints(self, ctx: ItemGetValueContext) -> str:
        with self.section_level_scope(ctx):
            content = SphinxContent(section_level=self.section_level,
                                    the_license=self.content_license)
            for item in self.spec.get_related_items_by_type("constraint"):
                self.add_item(content, item)
            return content.join()

    def _get_action_template(self, ctx: ItemGetValueContext) -> str:
        interface_functions = list(ctx.item.parents("interface-function"))
        if len(interface_functions) == 1:
            link = self.mapper.get_link(interface_functions[0])
            return f"When the {link} directive is called."
        return "When the directive is called."

    def _get_environment(self, _ctx: ItemGetValueContext) -> str:
        return "${ENVIRONMENT}"

    def _get_limit_kind(self, _ctx: ItemGetValueContext) -> str:
        return "${LIMIT_KIND}"

    def _get_limit_condition(self, _ctx: ItemGetValueContext) -> str:
        return "${LIMIT_CONDITION}"

    def _get_count(self, ctx: ItemGetValueContext) -> str:
        return str(ctx.value[ctx.key])

    def _add_requirements(self, content: SphinxContent, section: str,
                          types: tuple[str, ...]) -> None:
        with content.section(section):
            for item in self.spec.get_related_items_by_type(types):
                self.add_item(content, item)

    def _get_requirements(self, ctx: ItemGetValueContext) -> str:
        with self.section_level_scope(ctx):
            content = SphinxContent(section_level=self.section_level,
                                    the_license=self.content_license)
            content.add(".. include:: ../include/abbreviations.rst")
            with content.section("Requirements"):
                self._add_requirements(
                    content, "Functional requirements",
                    ("requirement/functional/action",
                     "requirement/functional/capability",
                     "requirement/functional/function",
                     "requirement/functional/interface-define-not-defined"))
                self._add_requirements(
                    content, "Performance requirements",
                    ("requirement/non-functional/performance",
                     "requirement/non-functional/performance-runtime",
                     "requirement/non-functional"
                     "/performance-runtime-environment",
                     "requirement/non-functional/performance-runtime-limits"))
                with content.section("Interface requirements"):
                    icd = self.substitute("${doc-ts-icd:/cite-long}")
                    content.add(f"For interface requirements see the {icd}.")
                with content.section("Operational requirements"):
                    _add_no_requirements(content, "operational")
                with content.section("Resources requirements"):
                    _add_no_requirements(content, "resources")
                self._add_requirements(
                    content,
                    "Design requirements and implementation constraints",
                    ("requirement/non-functional/design",
                     "requirement/non-functional/design-group",
                     "requirement/non-functional/design-target"))
                with content.section("Security and privacy requirements"):
                    _add_no_requirements(content, "security and privacy")
                with content.section("Portability requirements"):
                    _add_no_requirements(content, "portability")
                self._add_requirements(
                    content, "Software quality requirements",
                    ("requirement/non-functional/quality", ))
                with content.section("Software reliability requirements"):
                    _add_no_requirements(content, "software reliability")
                with content.section("Software maintainability requirements"):
                    _add_no_requirements(content, "software maintainability")
                with content.section("Software safety requirements"):
                    _add_no_requirements(content, "software safety")
                with content.section(
                        "Data definition and database requirements"):
                    _add_no_requirements(content,
                                         "data definition and database")
                with content.section("Human factors related requirements"):
                    _add_no_requirements(content, "human factors")
                with content.section(
                        "Adaptation and installation requirements"):
                    _add_no_requirements(content, "adaption and installation")
            return content.join()



# SPDX-License-Identifier: BSD-2-Clause
""" Provides a command line interface to generate interface items. """

# Copyright (C) 2024, 2025 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import contextlib
import os
import sys

from specitems import get_arguments, save_data
from specware import load_specware_config

from .sourcetospec import (doxygen_xml_to_spec, DoxygenContext, DoxygenGroup,
                           DoxygenFile, DoxygenFunction)


def _propose_config(ctx: DoxygenContext, spec_dir: str, config: dict) -> None:
    print("spec-from-source:")
    print(f"  spec-directory: {spec_dir}")
    print("  groups:")
    for group in sorted(ctx.items_by_kind["group"].values()):
        print(f"    {group.doxygen_id}: # {group.name} "
              f"({group.data['title']})")
        print(f"      uid: {group.uid}")
    for group in sorted(ctx.items_by_kind["group"].values()):
        assert isinstance(group, DoxygenGroup)
        if group.doxygen_id not in config["enabled-groups"]:
            continue
        print(group.doxygen_id)
        for group_member in group.members():
            if group_member.is_header:
                assert isinstance(group_member, DoxygenFile)
                print("  ", group_member.uid)
                for header_member in group_member.members():
                    if not isinstance(header_member, DoxygenFunction):
                        continue
                    path = os.path.join(spec_dir,
                                        f"{header_member.uid[1:]}.yml")
                    print("    ", header_member.uid, path)
                    save_data(path, header_member.export())


def clifromsource(argv: list[str] = sys.argv) -> None:
    """
    Generate interface items from Doxygen generated XML files using the
    configuration.
    """

    def _add_arguments(parser):
        parser.add_argument("--config-file",
                            type=str,
                            default=None,
                            help="use this configuration file")
        parser.add_argument("--propose-config",
                            action="store_true",
                            help="propose a configuration")
        parser.add_argument("doxygen_xml_files",
                            metavar="DOXYGEN_XML_FILES",
                            nargs="+",
                            help="the Doxygen generated XML files")

    args = get_arguments(argv[1:],
                         description=clifromsource.__doc__,
                         add_arguments=(_add_arguments, ))
    config, working_directory = load_specware_config(args.config_file)
    config = config["spec-from-source"]
    with contextlib.chdir(working_directory):
        spec_dir = config["spec-directory"]
        os.makedirs(spec_dir, exist_ok=True)
        ctx = doxygen_xml_to_spec(config, args.doxygen_xml_files)
        if args.propose_config:
            _propose_config(ctx, spec_dir, config)
        else:
            for item in ctx.items_by_kind["function"].values():
                print(item.name, item.data)



# SPDX-License-Identifier: BSD-2-Clause
""" Generates runtime performance measurement images. """

# Copyright (C) 2022, 2025 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import logging
import multiprocessing
import os
import matplotlib.pyplot as plt  # type: ignore
from matplotlib import ticker  # type: ignore

from specitems import Item

from .directorystate import DirectoryState
from .pkgitems import PackageBuildDirector
from .testaggregator import TestAggregator

logging.getLogger("matplotlib").setLevel(logging.WARNING)


def environment_order(env_name_data: dict) -> int:
    """
    Returns the environment order key for the environment name and data pair.
    """
    name = env_name_data[0]
    if name == "HotCache":
        return 0
    if name == "FullCache":
        return 1
    if name == "DirtyCache":
        return 2
    return int(name[5:]) + 2


def _worker(work_queue: multiprocessing.Queue, build_item_uid: str):
    while True:
        measurement_data = work_queue.get()
        if measurement_data is None:
            logging.info("%s: worker done", build_item_uid)
            return
        req_uid = measurement_data["requirement-uid"]
        config = measurement_data["test-case"]["test-suite"]["config"]
        config_key = config["config-key"]
        target_name = config["target"]["name"]
        logging.info("%s: create images for: %s - %s - %s", build_item_uid,
                     target_name, config_key, req_uid)
        spec = f"spec:{req_uid}"
        env_samples = []
        env_names = []
        max_median = 0
        for env_name, env_data in sorted(measurement_data["variants"].items(),
                                         key=environment_order):
            env_names.append(env_name)
            env_samples.append(env_data["samples"])
            max_median = max(max_median, env_data["q2"])
            base = env_data["histogram"]
            title = f"""{spec}
Target - {target_name}
Configuration - {config_key}
Measurement Environment - {env_name}"""
            _histogram(title, base, env_data)
        base = measurement_data["boxplot"]
        title = f"""{spec}
Target - {target_name}
Configuration - {config_key}"""
        _boxplot(title, base, max_median, env_samples, env_names)


def _scale_and_unit(value: float) -> tuple[float, str]:
    if value < 1e-6:
        return 1e9, "ns"
    if value < 1e-3:
        return 1e6, "μs"
    if value < 1.0:
        return 1e3, "ms"
    return 1.0, "s"


def _freedman_draconis(scale: float, env_data: dict) -> int:
    sample_count = len(env_data["samples"])
    try:
        width = 2 * scale * (env_data["q3"] -
                             env_data["q1"]) / sample_count**(1. / 3.)
        bin_count = int(scale * (env_data["max"] - env_data["min"]) / width)
        return min(bin_count, sample_count)
    except ZeroDivisionError:
        return sample_count


_PDF_METADATA = {"creationDate": None}


def _histogram(title: str, base: str, env_data: dict) -> None:
    scale, unit = _scale_and_unit(env_data["q2"])
    scaled_samples = list(scale * sample for sample in env_data["samples"])
    fig, axes = plt.subplots()
    axes.set_title(title)
    axes.set_xlabel(f"Runtime [{unit}]")
    axes.set_xlim(scale * env_data["min"], scale * env_data["max"])
    axes.set_ylabel("Sample Count")
    ymax = len(scaled_samples)
    ymin = 0
    axes.set_ylim(ymin, ymax)
    axes.hist(scaled_samples,
              _freedman_draconis(scale, env_data),
              color="deepskyblue")
    axes.hist(scaled_samples,
              list(sorted(set(scaled_samples))),
              histtype="step",
              color="black",
              cumulative=True)
    q123 = [
        scale * env_data["q1"], scale * env_data["q2"], scale * env_data["q3"]
    ]
    plt.vlines(q123, ymin=ymin, ymax=ymax)
    for i, quartile in enumerate(q123):
        plt.text(quartile, (2 * i + 1) * ymax / 8, f"Q{i + 1}")
    fig.tight_layout()
    plt.savefig(f"{base}.png")
    plt.savefig(f"{base}.pdf", metadata=_PDF_METADATA)
    plt.close()


def _boxplot(title: str, base: str, max_median: float,
             env_samples: list[list[float]], env_names: list[str]) -> None:
    scale, unit = _scale_and_unit(max_median)
    scaled_samples = []
    for samples in env_samples:
        scaled_samples.append(list(scale * sample for sample in samples))
    fig, axes = plt.subplots()
    axes.set_title(title)
    axes.set_xlabel("Measurement Environment")
    axes.xaxis.set_major_locator(
        ticker.FixedLocator(list(range(len(env_names), 1))))
    axes.xaxis.set_major_formatter(ticker.FixedFormatter(env_names))
    axes.set_ylabel(f"Runtime [{unit}]")
    axes.yaxis.set_major_formatter(ticker.FormatStrFormatter("%.3f"))
    axes.boxplot(scaled_samples)
    fig.tight_layout()
    plt.savefig(f"{base}.png")
    plt.savefig(f"{base}.pdf", metadata=_PDF_METADATA)
    plt.close()


class BuildPerformanceImages(DirectoryState):
    """ Generates runtime performance measurement images. """

    def __init__(self, director: PackageBuildDirector, item: Item):
        super().__init__(director, item)
        tests = self.input("test-aggregation")
        assert isinstance(tests, TestAggregator)
        files: list[str] = []
        for measurement_data in tests.runtime_measurements:
            uid = measurement_data["requirement-uid"]
            config = measurement_data["test-case"]["test-suite"]["config"]
            target = config["target"]
            name = f"{target['key']}-{config['config-key']}-{uid[1:]}".replace(
                "/", "-")
            for env_name, env_data in measurement_data["variants"].items():
                name_2 = f"{name}-{env_name.replace('/', '-')}"
                files.append(f"{name_2}.pdf")
                files.append(f"{name_2}.png")
                base = os.path.join(self.directory, name_2)
                env_data["histogram"] = base
            files.append(f"{name}.pdf")
            files.append(f"{name}.png")
            base = os.path.join(self.directory, name)
            measurement_data["boxplot"] = base
        self._perf_files = files

    def run(self) -> None:
        tests = self.input("test-aggregation")
        assert isinstance(tests, TestAggregator)
        self.discard()
        os.makedirs(self.directory, exist_ok=True)
        work_queue: multiprocessing.Queue = multiprocessing.Queue()
        for measurement_data in tests.runtime_measurements:
            work_queue.put(measurement_data)
        processes = []
        for _ in range(
                min(multiprocessing.cpu_count(),
                    len(tests.runtime_measurements))):
            process = multiprocessing.Process(target=_worker,
                                              args=(work_queue, self.uid),
                                              daemon=True)
            work_queue.put(None)
            process.start()
            processes.append(process)
        for process in processes:
            process.join()
        self.set_files(self._perf_files)

        self.description.add(f"""Produce performance images in directory
{self.description.path(self.directory)}""")



# SPDX-License-Identifier: BSD-2-Clause
"""
Provides a command line interface to create a package workspace directory
according to the configuration file.
"""

# Copyright (C) 2020, 2025 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import argparse
import logging
import os
import sys

from specware import run_command

from .pkgitems import BuildItem
from .pkgworkspace import (BuildspaceConfig, WorkspaceConfig, create_workspace,
                           export_to_buildspace)
from .util import get_build_arguments


def _get_arguments(argv: list[str]) -> argparse.Namespace:

    def _add_arguments(parser):
        parser.add_argument(
            "--do-not-use-git",
            help="do not use git to track changes in the workspace",
            action="store_true")
        parser.add_argument("--prefix",
                            help="the prefix directory",
                            default="/opt")
        parser.add_argument("--config-directory",
                            help="the configuration directory",
                            default=".")
        parser.add_argument("--enabled-set",
                            help="the enabled set to configure the package",
                            default="")
        parser.add_argument("--name", help="the package name", default="pkg")
        parser.add_argument("--cache-directory",
                            help="the configuration cache directory",
                            default="config-cache")
        parser.add_argument('config_files', nargs='+')

    return get_build_arguments(argv, add_arguments=(_add_arguments, ))


def _make_deployment_directory(workspace_package: BuildItem) -> str:
    deployment_directory = workspace_package["deployment-directory"]
    if os.path.exists(os.path.join(deployment_directory, ".git")):
        assert os.path.isdir(deployment_directory)
    else:
        logging.info("build: create deployment directory: %s",
                     deployment_directory)
        os.makedirs(deployment_directory, exist_ok=True)
        status = run_command(["git", "init"], deployment_directory)
        assert status == 0
    return deployment_directory


def clibuild(argv: list[str] = sys.argv) -> None:
    """
    Create a package workspace directory according to the configuration files.
    """

    args = _get_arguments(argv[1:])
    workspace_directory = os.path.abspath(args.config_directory)
    os.chdir(workspace_directory)

    logging.info("build: create workspace in: %s", workspace_directory)
    workspace_config = WorkspaceConfig(
        spec_directories=args.config_files,
        workspace_directory=workspace_directory,
        cache_directory=os.path.abspath(args.cache_directory),
        verify_specification_format=not args.no_spec_verify)
    workspace = create_workspace(workspace_config)

    deployment_directory = _make_deployment_directory(
        workspace.director.package)
    buildspace_config = BuildspaceConfig(
        spec_directory=os.path.join(deployment_directory, "build", "spec"),
        cache_directory=os.path.join(deployment_directory, "build", "cache"),
        use_git=not args.do_not_use_git)
    buildspace = export_to_buildspace(workspace, buildspace_config)
    buildspace.director.build_package(args.only, args.force, args.skip)



# SPDX-License-Identifier: BSD-2-Clause
""" Provides a test output parser. """

# Copyright (C) 2022, 2024 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import base64
import hashlib
import re
from typing import Any, Iterable

_TEST_BEGIN = re.compile(r"\*\*\* BEGIN OF TEST ([^*]+) \*\*\*")
_TEST_VERSION = re.compile(r"\*\*\* TEST VERSION: (.+)")
_TEST_STATE = re.compile(r"\*\*\* TEST STATE: (.+)")
_TEST_BUILD = re.compile(r"\*\*\* TEST BUILD: ?(.*)")
_TEST_TOOLS = re.compile(r"\*\*\* TEST TOOLS: (.+)")
_TEST_END = re.compile(r"\*\*\* END OF TEST ([^*]+) \*\*\*")

_TS_SUITE_BEGIN = re.compile(r"A:([^:]+)")
_TS_SUITE_END = re.compile(
    r"Z:([^:]+):C:([^:]+):N:([^:]+):F:([^:]+):D:([^:]+)")
_TS_CASE_BEGIN = re.compile(r"B:([^:]+)")
_TS_CASE_END = re.compile(r"E:([^:]+):N:([^:]+):F:([^:]+):D:([^:]+)")
_TS_PLATFORM = re.compile(r"S:Platform:(.+)")
_TS_COMPILER = re.compile(r"S:Compiler:(.+)")
_TS_VERSION = re.compile(r"S:Version:(.+)")
_TS_BSP = re.compile(r"S:BSP:(.+)")
_TS_BUILD_LABEL = re.compile(r"S:BuildLabel:(.+)")
_TS_TARGET_HASH = re.compile(r"S:TargetHash:SHA256:(.*)")
_TS_RTEMS_DEBUG = re.compile(r"S:RTEMS_DEBUG:([01])$")
_TS_RTEMS_MULTIPROCESSING = re.compile(r"S:RTEMS_MULTIPROCESSING:([01])$")
_TS_RTEMS_POSIX_API = re.compile(r"S:RTEMS_POSIX_API:([01])$")
_TS_RTEMS_PROFILING = re.compile(r"S:RTEMS_PROFILING:([01])$")
_TS_RTEMS_SMP = re.compile(r"S:RTEMS_SMP:([01])$")
_TS_REMARK = re.compile(r"R:(.+)")
_TS_REPORT_HASH = re.compile(r"Y:ReportHash:SHA256:(.+)")

_M_BEGIN = re.compile(r"M:B:(.+)")
_M_END = re.compile(r"M:E:([^:]+):D:(.+)")
_M_V = re.compile(r"M:V:(.+)")
_M_N = re.compile(r"M:N:(.+)")
_M_S = re.compile(r"M:S:([^:]+):(.+)")
_M_MI = re.compile(r"M:MI:(.+)")
_M_P1 = re.compile(r"M:P1:(.+)")
_M_Q1 = re.compile(r"M:Q1:(.+)")
_M_Q2 = re.compile(r"M:Q2:(.+)")
_M_Q3 = re.compile(r"M:Q3:(.+)")
_M_P99 = re.compile(r"M:P99:(.+)")
_M_MX = re.compile(r"M:MX:(.+)")
_M_MAD = re.compile(r"M:MAD:(.+)")
_M_D = re.compile(r"M:D:(.+)")

_GCOV_BEGIN = "*** BEGIN OF GCOV INFO BASE64 ***"
_GCOV_END = "*** END OF GCOV INFO BASE64 ***"
_GCOV_HASH = re.compile(r"\*\*\* GCOV INFO SHA256 (.*) \*\*\*")

_RECORDS_BEGIN = "*** BEGIN OF RECORDS BASE64 ***"
_RECORDS_END = "*** END OF RECORDS BASE64 ***"
_RECORDS_ZLIB_BEGIN = "*** BEGIN OF RECORDS BASE64 ZLIB ***"
_RECORDS_ZLIB_END = "*** END OF RECORDS BASE64 ZLIB ***"


def _are_samples_valid(measurement) -> bool:
    if len(measurement["samples"]) != measurement["sample-count"]:
        return False
    if not measurement["samples"]:
        return True
    if measurement["min"] != measurement["samples"][0]:
        return False
    return measurement["max"] == measurement["samples"][-1]


class TestOutputParser:
    """ Provides a line by line parser of test output. """

    # pylint: disable=too-few-public-methods
    # pylint: disable=too-many-instance-attributes
    def __init__(self, data) -> None:
        self.data = data
        self.consume = self._test_begin
        self._consume_previous = self.consume
        self.hash_line = self._hash_none
        assert "info" not in data
        self.data["info"] = {}
        self.data["data-ranges"] = []
        assert "test-suite" not in data
        self.level = 0
        self._hash_state = hashlib.sha256()
        self._measurement: dict[str, Any] = {}
        self._test_case: dict[str, Any] = {}

    def _error(self, index: int) -> bool:
        if "line-parser-error" not in self.data:
            self.data["line-parser-error"] = index
        self.consume = self._extra
        return False

    def _hash_none(self, line: str) -> None:
        pass

    def _hash_sha256(self, line: str) -> None:
        self._hash_state.update(f"{line}\n".encode("latin-1"))

    def _hash_sha256_skip_one(self, line: str) -> None:
        # pylint: disable=unused-argument
        self.hash_line = self._hash_sha256

    def _hash_finalize(self) -> str:
        self.hash_line = self._hash_none
        digest = base64.urlsafe_b64encode(
            self._hash_state.digest()).decode("ascii")
        return digest

    def _test_begin(self, index: int, line: str) -> bool:
        mobj = _TEST_BEGIN.match(line)
        if mobj:
            self.level += 1
            self.data["info"]["line-begin-of-test"] = index
            self.data["info"]["name"] = mobj.group(1)
            self.consume = self._test_version
            return True
        return self._extra(index, line)

    def _test_end(self, index: int, line: str) -> bool:
        mobj = _TEST_END.match(line)
        if mobj:
            self.level -= 1
            if self.data["info"].get("name", None) == mobj.group(1):
                self.data["info"]["line-end-of-test"] = index
                return True
            return self._error(index)
        return False

    def _test_version(self, index: int, line: str) -> bool:
        mobj = _TEST_VERSION.match(line)
        if mobj:
            self.data["info"]["version"] = mobj.group(1)
            self.data["info"]["line-version"] = index
            self.consume = self._test_state
            return True
        self.consume = self._test_body
        return self._test_body(index, line)

    def _test_state(self, index: int, line: str) -> bool:
        mobj = _TEST_STATE.match(line)
        if mobj:
            self.data["info"]["state"] = mobj.group(1)
            self.data["info"]["line-state"] = index
            self.consume = self._test_build
            return True
        return self._error(index)

    def _test_build(self, index: int, line: str) -> bool:
        mobj = _TEST_BUILD.match(line)
        if mobj:
            build = mobj.group(1)
            if build:
                self.data["info"]["build"] = build.split(", ")
            else:
                self.data["info"]["build"] = []
            self.data["info"]["line-build"] = index
            self.consume = self._test_tools
            return True
        return self._error(index)

    def _test_tools(self, index: int, line: str) -> bool:
        mobj = _TEST_TOOLS.match(line)
        if mobj:
            self.data["info"]["tools"] = mobj.group(1)
            self.data["info"]["line-tools"] = index
            self.consume = self._test_body
            return True
        return self._error(index)

    def _test_body(self, index: int, line: str) -> bool:
        if self._test_suite_begin(index, line):
            return True
        return self._extra(index, line)

    def _test_suite_begin(self, index: int, line: str) -> bool:
        mobj = _TS_SUITE_BEGIN.match(line)
        if mobj:
            self.level += 1
            self.data["test-suite"] = {
                "duration": "?",
                "failed-steps-count": "?",
                "line-begin": index,
                "line-duration": "?",
                "line-end": "?",
                "line-failed-steps-count": "?",
                "line-report-hash": "?",
                "line-step-count": "?",
                "name": mobj.group(1),
                "report-hash": "?",
                "report-hash-calculated": "?",
                "step-count": "?",
                "test-cases": []
            }
            self.consume = self._test_suite_platform
            self._hash_state = hashlib.sha256()
            self.hash_line = self._hash_sha256
            return True
        return False

    def _test_suite_platform(self, index: int, line: str) -> bool:
        mobj = _TS_PLATFORM.match(line)
        if mobj:
            self.data["test-suite"]["platform"] = mobj.group(1)
            self.data["test-suite"]["line-platform"] = index
            self.consume = self._test_suite_compiler
            return True
        return self._error(index)

    def _test_suite_compiler(self, index: int, line: str) -> bool:
        mobj = _TS_COMPILER.match(line)
        if mobj:
            self.data["test-suite"]["compiler"] = mobj.group(1)
            self.data["test-suite"]["line-compiler"] = index
            self.consume = self._test_suite_version
            return True
        return self._error(index)

    def _test_suite_version(self, index: int, line: str) -> bool:
        mobj = _TS_VERSION.match(line)
        if mobj:
            self.data["test-suite"]["version"] = mobj.group(1)
            self.data["test-suite"]["line-version"] = index
            self.consume = self._test_suite_bsp
            return True
        return self._error(index)

    def _test_suite_bsp(self, index: int, line: str) -> bool:
        mobj = _TS_BSP.match(line)
        if mobj:
            self.data["test-suite"]["bsp"] = mobj.group(1)
            self.data["test-suite"]["line-bsp"] = index
            self.consume = self._test_suite_build_label
            return True
        return self._error(index)

    def _test_suite_build_label(self, index: int, line: str) -> bool:
        mobj = _TS_BUILD_LABEL.match(line)
        if mobj:
            self.data["test-suite"]["build-label"] = mobj.group(1)
            self.data["test-suite"]["line-build-label"] = index
            self.consume = self._test_suite_target_hash
            return True
        return self._error(index)

    def _test_suite_target_hash(self, index: int, line: str) -> bool:
        mobj = _TS_TARGET_HASH.match(line)
        if mobj:
            self.data["test-suite"]["target-hash"] = mobj.group(1)
            self.data["test-suite"]["line-target-hash"] = index
            self.consume = self._test_suite_rtems_debug
            return True
        return self._error(index)

    def _test_suite_rtems_debug(self, index: int, line: str) -> bool:
        mobj = _TS_RTEMS_DEBUG.match(line)
        if mobj:
            self.data["test-suite"]["rtems-debug"] = bool(int(mobj.group(1)))
            self.data["test-suite"]["line-rtems-debug"] = index
            self.consume = self._test_suite_rtems_multiprocessing
            return True
        return self._error(index)

    def _test_suite_rtems_multiprocessing(self, index: int, line: str) -> bool:
        mobj = _TS_RTEMS_MULTIPROCESSING.match(line)
        if mobj:
            self.data["test-suite"]["rtems-multiprocessing"] = bool(
                int(mobj.group(1)))
            self.data["test-suite"]["line-rtems-multiprocessing"] = index
            self.consume = self._test_suite_rtems_posix_api
            return True
        return self._error(index)

    def _test_suite_rtems_posix_api(self, index: int, line: str) -> bool:
        mobj = _TS_RTEMS_POSIX_API.match(line)
        if mobj:
            self.data["test-suite"]["rtems-posix-api"] = bool(
                int(mobj.group(1)))
            self.data["test-suite"]["line-rtems-posix-api"] = index
            self.consume = self._test_suite_rtems_profiling
            return True
        return self._error(index)

    def _test_suite_rtems_profiling(self, index: int, line: str) -> bool:
        mobj = _TS_RTEMS_PROFILING.match(line)
        if mobj:
            self.data["test-suite"]["rtems-profiling"] = bool(
                int(mobj.group(1)))
            self.data["test-suite"]["line-rtems-profiling"] = index
            self.consume = self._test_suite_rtems_smp
            return True
        return self._error(index)

    def _test_suite_rtems_smp(self, index: int, line: str) -> bool:
        mobj = _TS_RTEMS_SMP.match(line)
        if mobj:
            self.data["test-suite"]["rtems-smp"] = bool(int(mobj.group(1)))
            self.data["test-suite"]["line-rtems-smp"] = index
            self.consume = self._test_suite_body
            return True
        return self._error(index)

    def _test_suite_body(self, index: int, line: str) -> bool:
        if self._test_case_begin(index, line):
            return True
        mobj = _TS_SUITE_END.match(line)
        if mobj:
            self.level -= 1
            data = self.data["test-suite"]
            count = int(mobj.group(2))
            if data["name"] == mobj.group(1) and len(
                    data["test-cases"]) == count:
                data["line-end"] = index
                data["line-step-count"] = index
                data["line-failed-steps-count"] = index
                data["line-duration"] = index
                data["step-count"] = int(mobj.group(3))
                data["failed-steps-count"] = int(mobj.group(4))
                data["duration"] = float(mobj.group(5))
                self.consume = self._report_hash
                return True
            return self._error(index)
        return self._extra(index, line)

    def _test_case_begin(self, index: int, line: str) -> bool:
        mobj = _TS_CASE_BEGIN.match(line)
        if mobj:
            self.level += 1
            self._test_case = {
                "line-begin": index,
                "name": mobj.group(1),
                "remarks": [],
                "runtime-measurements": []
            }
            self.consume = self._test_case_body
            return True
        return self._extra(index, line)

    def _remark(self, index: int, line: str) -> bool:
        mobj = _TS_REMARK.match(line)
        if mobj:
            self._test_case["remarks"].append({
                "line": index,
                "remark": mobj.group(1)
            })
            return True
        return False

    def _test_case_body(self, index: int, line: str) -> bool:
        if self._measurement_begin(index, line):
            return True
        if self._remark(index, line):
            return True
        mobj = _TS_CASE_END.match(line)
        if mobj:
            self.level -= 1
            if self._test_case["name"] == mobj.group(1):
                self._test_case["line-end"] = index
                self._test_case["line-step-count"] = index
                self._test_case["line-failed-steps-count"] = index
                self._test_case["line-duration"] = index
                self._test_case["step-count"] = int(mobj.group(2))
                self._test_case["failed-steps-count"] = int(mobj.group(3))
                self._test_case["duration"] = float(mobj.group(4))
                self.data["test-suite"]["test-cases"].append(self._test_case)
                self.consume = self._test_suite_body
                return True
            return self._error(index)
        return self._extra(index, line)

    def _measurement_begin(self, index: int, line: str) -> bool:
        mobj = _M_BEGIN.match(line)
        if mobj:
            self.level += 1
            self._measurement = {
                "line-begin": index,
                "name": mobj.group(1),
                "samples": []
            }
            self.consume = self._measurement_variant
            return True
        return False

    def _measurement_variant(self, index: int, line: str) -> bool:
        mobj = _M_V.match(line)
        if mobj:
            self._measurement["variant"] = mobj.group(1)
            self.consume = self._measurement_count
            return True
        return self._error(index)

    def _measurement_count(self, index: int, line: str) -> bool:
        mobj = _M_N.match(line)
        if mobj:
            self._measurement["sample-count"] = int(mobj.group(1))
            self.consume = self._measurement_samples
            return True
        return self._error(index)

    def _measurement_samples(self, index: int, line: str) -> bool:
        if self._measurement_min(index, line):
            return True
        mobj = _M_S.match(line)
        if mobj:
            self._measurement["samples"].extend(  # type: ignore
                int(mobj.group(1)) * [float(mobj.group(2))])
            return True
        return self._error(index)

    def _measurement_min(self, index: int, line: str) -> bool:
        mobj = _M_MI.match(line)
        if mobj:
            self._measurement["min"] = float(mobj.group(1))
            self.consume = self._measurement_p1
            return True
        return self._extra(index, line)

    def _measurement_p1(self, index: int, line: str) -> bool:
        mobj = _M_P1.match(line)
        if mobj:
            self._measurement["p1"] = float(mobj.group(1))
            self.consume = self._measurement_q1
            return True
        return self._error(index)

    def _measurement_q1(self, index: int, line: str) -> bool:
        mobj = _M_Q1.match(line)
        if mobj:
            self._measurement["q1"] = float(mobj.group(1))
            self.consume = self._measurement_q2
            return True
        return self._error(index)

    def _measurement_q2(self, index: int, line: str) -> bool:
        mobj = _M_Q2.match(line)
        if mobj:
            self._measurement["q2"] = float(mobj.group(1))
            self.consume = self._measurement_q3
            return True
        return self._error(index)

    def _measurement_q3(self, index: int, line: str) -> bool:
        mobj = _M_Q3.match(line)
        if mobj:
            self._measurement["q3"] = float(mobj.group(1))
            self.consume = self._measurement_p99
            return True
        return self._error(index)

    def _measurement_p99(self, index: int, line: str) -> bool:
        mobj = _M_P99.match(line)
        if mobj:
            self._measurement["p99"] = float(mobj.group(1))
            self.consume = self._measurement_max
            return True
        return self._error(index)

    def _measurement_max(self, index: int, line: str) -> bool:
        mobj = _M_MX.match(line)
        if mobj:
            self._measurement["max"] = float(mobj.group(1))
            self.consume = self._measurement_mad
            return True
        return self._error(index)

    def _measurement_mad(self, index: int, line: str) -> bool:
        mobj = _M_MAD.match(line)
        if mobj:
            self._measurement["mad"] = float(mobj.group(1))
            self.consume = self._measurement_duration
            return True
        return self._error(index)

    def _measurement_duration(self, index: int, line: str) -> bool:
        mobj = _M_D.match(line)
        if mobj:
            self._measurement["duration-sum"] = float(mobj.group(1))
            self.consume = self._measurement_end
            return True
        return self._error(index)

    def _measurement_end(self, index: int, line: str) -> bool:
        mobj = _M_END.match(line)
        if mobj:
            self.level -= 1
            if self._measurement["name"] == mobj.group(
                    1) and _are_samples_valid(self._measurement):
                self._measurement["line-end"] = index
                self._measurement["duration-total"] = float(mobj.group(2))
                self._test_case["runtime-measurements"].append(  # type: ignore
                    self._measurement)
                self.consume = self._test_case_body
                return True
        return self._error(index)

    def _report_hash(self, index: int, line: str) -> bool:
        mobj = _TS_REPORT_HASH.match(line)
        if mobj:
            self.data["test-suite"][
                "report-hash-calculated"] = self._hash_finalize()
            self.data["test-suite"]["report-hash"] = mobj.group(1)
            self.data["test-suite"]["line-report-hash"] = index
            self.consume = self._test_body
            return True
        return self._extra(index, line)

    def _gcov_begin(self, index: int, line: str) -> bool:
        if line == _GCOV_BEGIN:
            self.level += 1
            self.data["line-gcov-info-base64-begin"] = index
            self._consume_previous = self.consume
            self.consume = self._gcov_end
            self._hash_state = hashlib.sha256()
            self.hash_line = self._hash_sha256_skip_one
            return True
        return False

    def _gcov_end(self, index: int, line: str) -> bool:
        if line == _GCOV_END:
            self.level -= 1
            self.data["gcov-info-hash-calculated"] = self._hash_finalize()
            self.data["line-gcov-info-base64-end"] = index
            self.data["data-ranges"].append(
                (self.data["line-gcov-info-base64-begin"] + 1, index))
            self.consume = self._consume_previous
            return True
        return False

    def _gcov_hash(self, index: int, line: str) -> bool:
        mobj = _GCOV_HASH.match(line)
        if mobj:
            self.data["gcov-info-hash"] = mobj.group(1)
            self.data["line-gcov-info-hash"] = index
            return True
        return False

    def _records_begin(self, index: int, line: str) -> bool:
        if line == _RECORDS_BEGIN:
            self.level += 1
            self.data["line-records-base64-begin"] = index
            self._consume_previous = self.consume
            self.consume = self._records_end
            return True
        return False

    def _records_end(self, index: int, line: str) -> bool:
        if line == _RECORDS_END:
            self.level -= 1
            self.data["line-records-base64-end"] = index
            self.data["data-ranges"].append(
                (self.data["line-records-base64-begin"] + 1, index))
            self.consume = self._consume_previous
            return True
        return False

    def _records_zlib_begin(self, index: int, line: str) -> bool:
        if line == _RECORDS_ZLIB_BEGIN:
            self.level += 1
            self.data["line-records-base64-zlib-begin"] = index
            self._consume_previous = self.consume
            self.consume = self._records_zlib_end
            return True
        return False

    def _records_zlib_end(self, index: int, line: str) -> bool:
        if line == _RECORDS_ZLIB_END:
            self.level -= 1
            self.data["line-records-base64-zlib-end"] = index
            self.data["data-ranges"].append(
                (self.data["line-records-base64-zlib-begin"] + 1, index))
            self.consume = self._consume_previous
            return True
        return False

    def _extra(self, index: int, line: str) -> bool:
        if self._test_end(index, line):
            return True
        if self._gcov_begin(index, line):
            return True
        if self._gcov_hash(index, line):
            return True
        if self._records_begin(index, line):
            return True
        if self._records_zlib_begin(index, line):
            return True
        return False


def augment_report(report: dict[str, Any], output: Iterable[str]) -> None:
    """ Augment the report with the results of the parsed output. """
    test_parser = TestOutputParser(report)
    for index, line in enumerate(output):
        if not line:
            continue
        test_parser.consume(index, line)
        test_parser.hash_line(line)



# SPDX-License-Identifier: BSD-2-Clause
""" Provides support to populate a workspace. """

# Copyright (C) 2025, 2026 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import contextlib
import copy
import dataclasses
import itertools
import graphlib
import logging
import os
import shutil
import subprocess
from typing import Iterator, Optional

from specitems import (EnabledSet, Item, ItemCache, ItemCacheConfig,
                       ItemDataByUID, ItemGetValueContext, ItemSelection,
                       JSONItemCache, data_digest, hash_file, is_enabled,
                       link_is_enabled, load_data, to_iterable,
                       verify_specification_format)
from specware import run_command

from .directorystate import DirectoryState, RepositoryState
from .pkgfactory import create_build_item_factory
from .pkgitems import (BuildItem, BuildItemFactory, BuildItemMapper,
                       BuildItemTypeProvider, PackageBuildDirector,
                       PackageComponent)
from .pkgtemplate import (ComponentTemplate, FileItemTemplate,
                          InlineItemTemplate)


def _verify_specification_format(item_cache: ItemCache, verify: bool):
    if verify:
        logger = logging.getLogger()
        level = logger.getEffectiveLevel()
        logger.setLevel(logging.ERROR)
        status = verify_specification_format(item_cache)
        assert status.critical == 0
        assert status.error == 0
        logger.setLevel(level)


def _workspace_hash(data: dict) -> str:
    # Filter out link hashes and the workspace hash
    links = []
    for link in data["links"]:
        if "hash" in link:
            link_2 = link.copy()
            link_2["hash"] = None
            links.append(link_2)
        else:
            links.append(link)
    data_2 = data.copy()
    data_2["links"] = links
    data_2["workspace-hash"] = None
    return data_digest(data_2)


_DUMMY_DATA = {
    "SPDX-License-Identifier": "CC-BY-SA-4.0 OR BSD-2-Clause",
    "copyrights": ["Copyright (C) 2025 embedded brains GmbH & Co. KG"],
    "enabled-by": False,
    "links": [],
    "pkg-type": "dummy",
    "type": "pkg",
    "workspace-hash": None
}

_DUMMY_DATA["workspace-hash"] = _workspace_hash(_DUMMY_DATA)

_DEFAULT_ATTRIBUTES = ("SPDX-License-Identifier", "copyrights", "enabled-by",
                       "links", "type")


def _make_directory_state_data(item: Item, directory_state_type: str) -> dict:
    return {
        "directory":
        item["destination-directory"],
        "directory-state-type":
        item.get("directory-state-type", directory_state_type),
        "files": [],
        "hash":
        None,
        "pkg-type":
        "directory-state"
    }


class _WorkspaceItem(BuildItem):

    @contextlib.contextmanager
    def selection_and_view_scope(self) -> Iterator[None]:
        """ Opens a selection and view context. """
        if isinstance(self, WorkspaceComponent):
            workspace_component: PackageComponent = self
        else:
            workspace_component = self.component
        item_cache = self.director.item_cache
        item_cache.push_selection(workspace_component.selection)
        item_cache.push_view(workspace_component.view)
        yield
        item_cache.pop_view()
        item_cache.pop_selection()

    def load_workspace_state(self) -> str:
        """
        Load the workspace state of the item.  Returns the hash of the state.
        """
        return _workspace_hash(self.item.data)

    def copy_attributes(self, data: dict, attributes: tuple[str, ...]) -> None:
        """
        Assign a deep copy of default attributes and the specified attributes
        of the item to the associated data elements.
        """
        for key in itertools.chain(_DEFAULT_ATTRIBUTES, attributes):
            try:
                value = copy.deepcopy(self.item[key])
            except KeyError:
                pass
            else:
                data[key] = value

    def get_buildspace_data(self) -> dict:
        """
        Export the workspace item state from the workspace to the buildspace.
        Return the buildspace item data.
        """
        raise NotImplementedError

    def copy_to_buildspace(self, buildspace_item: BuildItem) -> None:
        """ Copies the associated workspace state to the buildspace. """


def _apply_patches(workspace_directory: str,
                   unpacked_archive: DirectoryState) -> None:
    for patch in unpacked_archive["archive-patches"]:
        command = [
            "git", "--git-dir=.", "apply", "-v", "-p", "1",
            os.path.join(workspace_directory, patch["file"])
        ]
        env = copy.deepcopy(os.environ.copy())
        env["LANG"] = "en_US.UTF-8"
        logging.info("%s: apply patch in '%s': %s", unpacked_archive.uid,
                     unpacked_archive.directory, " ".join(command))
        output = subprocess.check_output(command,
                                         cwd=unpacked_archive.directory,
                                         env=env,
                                         stderr=subprocess.STDOUT)
        patched_files = [
            line[14:].removesuffix(" cleanly.")
            for line in output.decode("utf-8").splitlines()
            if line.startswith("Applied patch ")
        ]
        logging.info("%s: patched files: %s", unpacked_archive.uid,
                     patched_files)
        unpacked_archive.add_files(patched_files)


class WorkspaceArchive(_WorkspaceItem):
    """ Represents a workspace archive. """

    def get_buildspace_data(self) -> dict:
        data = _make_directory_state_data(self.item, "unpacked-archive")
        data["archive-file"] = os.path.basename(self["archive-file"])
        self.copy_attributes(
            data, ("archive-hash", "archive-patches", "archive-url",
                   "copyrights-by-license", "description", "symbolic-links"))
        return data

    def copy_to_buildspace(self, buildspace_item: BuildItem) -> None:
        assert isinstance(buildspace_item, DirectoryState)
        archive_file = self["archive-file"]
        assert hash_file(archive_file) == buildspace_item["archive-hash"]
        buildspace_item.add_tarfile_members(archive_file,
                                            buildspace_item.directory, True)
        buildspace_item.compact()
        buildspace_item.create_symbolic_links(
            buildspace_item["symbolic-links"])
        _apply_patches(self.substitute("${.:/component/workspace-directory}"),
                       buildspace_item)
        buildspace_item.load()


def _get_view(ctx: ItemGetValueContext) -> str:
    return ctx.item.view[ctx.key]


class WorkspaceComponent(_WorkspaceItem, PackageComponent):
    """ Represents a workspace component. """

    @classmethod
    def prepare_factory(cls, factory: BuildItemFactory,
                        type_name: str) -> None:
        PackageComponent.prepare_factory(factory, type_name)
        factory.add_get_value(f"{type_name}:/workspace-commit", _get_view)
        factory.add_get_value(f"{type_name}:/workspace-directory", _get_view)

    def get_buildspace_data(self) -> dict:
        data = copy.deepcopy(self.item.data)
        del data["enabled-set-actions"]
        del data["enabled-set-feedback-mapping"]
        del data["workspace-type"]
        data["pkg-type"] = "component"
        return data


class WorkspaceDirectory(_WorkspaceItem, DirectoryState):
    """ Represents a workspace directory. """

    def load_workspace_state(self) -> str:
        self.load()
        return super().load_workspace_state()

    def get_buildspace_data(self) -> dict:
        data = _make_directory_state_data(self.item, "explicit")
        self.copy_attributes(data, ("directory-state-load-before-use",
                                    "copyrights-by-license", "params"))
        return data

    def _copy_to_buildspace_actions(self,
                                    buildspace_item: DirectoryState) -> None:
        buildspace_item.run_actions_provider.run(
            self.item.get("copy-to-buildspace-actions", {}))

    def copy_to_buildspace(self, buildspace_item: BuildItem) -> None:
        assert isinstance(buildspace_item, DirectoryState)
        buildspace_item.lazy_clone(self,
                                   file_mappings=self.item.get(
                                       "file-mappings", None))
        self._copy_to_buildspace_actions(buildspace_item)


class WorkspaceFile(WorkspaceDirectory):
    """ Represents a workspace file. """

    def load_workspace_state(self) -> str:
        for source_file in to_iterable(self["source-file"]):
            source_path = os.path.join(self.directory, source_file)
            if os.path.isfile(source_path):
                self.set_files([source_file])
                break
            logging.info("%s: source file does not exist: %s", self.uid,
                         source_path)
        else:
            raise ValueError(f"{self.uid}: no source file available")
        return super().load_workspace_state()

    def copy_to_buildspace(self, buildspace_item: BuildItem) -> None:
        assert isinstance(buildspace_item, DirectoryState)
        buildspace_item.copy_file(self.file, self["destination-file"])
        buildspace_item.load()
        self._copy_to_buildspace_actions(buildspace_item)


class WorkspaceTestLog(WorkspaceFile):
    """ Represents a workspace test log. """

    def get_buildspace_data(self) -> dict:
        data = super().get_buildspace_data()
        data["directory-state-type"] = "test-log"
        return data


class WorkspaceRedirect(_WorkspaceItem):
    """ Represents a workspace item redirection. """

    def load_workspace_state(self) -> str:
        data = load_data(self["source"])
        data["links"][:0] = self.item["links"]
        self.item["redirect-data"] = data
        return super().load_workspace_state()

    def get_buildspace_data(self) -> dict:
        return self.item["redirect-data"]


def _git_commit(repository_directory: str,
                commit: Optional[str] = None) -> str:
    if commit is None:
        commit = "HEAD"
    stdout: list[str] = []
    status = run_command(["git", "rev-parse", commit], repository_directory,
                         stdout)
    assert status == 0
    return stdout[0].strip()


class WorkspaceRepository(_WorkspaceItem):
    """ Represents a workspace repository. """

    def __init__(self,
                 director: PackageBuildDirector,
                 item: Item,
                 mapper: Optional[BuildItemMapper] = None) -> None:
        super().__init__(director, item, mapper)
        my_type = self.item.type
        self.mapper.add_get_value(f"{my_type}:/git-commit", self._git_commit)
        self.mapper.add_get_value(f"{my_type}:/git-commit-head",
                                  self._git_commit_head)

    def load_workspace_state(self) -> str:
        self.director.package.view.setdefault("submodules",
                                              []).append(self.uid)
        self.item["commit"] = self["commit"]
        self.item["origin-commit"] = self["origin-commit"]
        return super().load_workspace_state()

    def get_buildspace_data(self) -> dict:
        data = _make_directory_state_data(self.item, "repository")
        data["commit"] = self["commit"]
        data["origin-commit"] = self["origin-commit"]
        self.copy_attributes(
            data,
            ("branch", "copyrights-by-license", "description", "origin-branch",
             "origin-commit", "origin-commit-url", "origin-url"))
        return data

    def copy_to_buildspace(self, buildspace_item: BuildItem) -> None:
        assert isinstance(buildspace_item, RepositoryState)
        destination_directory = buildspace_item.directory
        logging.info("%s: remove Git repository '%s'", self.uid,
                     destination_directory)
        try:
            shutil.rmtree(destination_directory)
        except FileNotFoundError:
            assert not os.path.exists(destination_directory)
        self._git_clone(buildspace_item)

    def _git_commit(self, ctx: ItemGetValueContext) -> str:
        return _git_commit(self["directory"], ctx.args)

    def _git_commit_head(self, _ctx: ItemGetValueContext) -> str:
        return _git_commit(self["directory"])

    def _git_clone(self, buildspace_item: RepositoryState) -> None:
        source_directory = self["directory"]
        branch = self["branch"]
        commit = self["commit"]
        status = run_command(["git", "branch", "-f", branch, commit],
                             source_directory)
        if status != 0:
            status = run_command(["git", "fetch", "origin"], source_directory)
            assert status == 0
            status = run_command(["git", "branch", "-f", branch, commit],
                                 source_directory)
            assert status == 0
        destination_directory = buildspace_item.directory
        clone_command = [
            "git", "clone", "--no-local", "--branch", branch, "--single-branch"
        ]
        clone_depth = self["clone-depth"]
        if clone_depth is not None:
            clone_command.extend(["--depth", str(clone_depth)])
        clone_command.extend(
            [f"file://{source_directory}", destination_directory])
        status = run_command(clone_command, source_directory)
        assert status == 0
        for fetch in self["origin-fetch"]:
            if is_enabled(self.enabled_set, fetch["enabled-by"]):
                status = run_command(
                    ["git", "fetch", "origin", fetch["value"]],
                    destination_directory)
                assert status == 0
        origin_url = self["origin-url"]
        if origin_url:
            status = run_command(["git", "remote", "add", "tmp", origin_url],
                                 destination_directory)
            assert status == 0
            status = run_command(["git", "remote", "remove", "origin"],
                                 destination_directory)
            assert status == 0
            status = run_command(["git", "remote", "rename", "tmp", "origin"],
                                 destination_directory)
            assert status == 0
        origin_branch = self["origin-branch"]
        origin_commit = self["origin-commit"]
        if origin_branch and origin_commit:
            status = run_command(
                ["git", "checkout", "-b", origin_branch, origin_commit],
                destination_directory)
            assert status == 0
            status = run_command(["git", "checkout", branch],
                                 destination_directory)
            assert status == 0
            status = run_command([
                "git", "symbolic-ref", "refs/remotes/origin/HEAD",
                f"refs/remotes/origin/{origin_branch}"
            ], destination_directory)
            assert status == 0
        buildspace_item.create_symbolic_links(
            buildspace_item.substitute(self.item["symbolic-links"]))
        for command in self["post-clone-commands"]:
            status = run_command(command, destination_directory)
            assert status == 0
        _git_commit(destination_directory)
        buildspace_item.item["patterns"] = [{"include": "**/*", "exclude": []}]
        buildspace_item.load()
        del buildspace_item.item.data["patterns"]
        buildspace_item.git_add(destination_directory)


def create_workspace_item_factory() -> BuildItemFactory:
    """ Create the workspace build item factory. """
    factory = BuildItemFactory()
    factory.add_constructor("pkg/template/component", ComponentTemplate)
    factory.add_constructor("pkg/template/file-item", FileItemTemplate)
    factory.add_constructor("pkg/template/inline-item", InlineItemTemplate)
    factory.add_constructor("pkg/workspace/archive", WorkspaceArchive)
    factory.add_constructor("pkg/workspace/component", WorkspaceComponent)
    factory.add_constructor("pkg/workspace/directory", WorkspaceDirectory)
    factory.add_constructor("pkg/workspace/file", WorkspaceFile)
    factory.add_constructor("pkg/workspace/redirect", WorkspaceRedirect)
    factory.add_constructor("pkg/workspace/repository", WorkspaceRepository)
    factory.add_constructor("pkg/workspace/test-log", WorkspaceTestLog)
    return factory


def _apply_enabled_set_actions(component: WorkspaceComponent,
                               selection: ItemSelection, item: Item) -> None:
    for action in component.substitute(item["enabled-set-actions"]):
        selection.apply_action(action)
    logging.info("%s: enabled set after %s actions: %s", component.uid,
                 item.uid, sorted(selection.enabled_set))
    for template in itertools.chain(
            item.parents("use-package-component-template"),
            item.children("add-package-component-template")):
        _apply_enabled_set_actions(component, selection, template)


def _gather_feedback(component: WorkspaceComponent, selection: ItemSelection,
                     item: Item, feedback_set: set[str]) -> None:
    for template in itertools.chain(
            item.parents("use-package-component-template"),
            item.children("add-package-component-template")):
        _gather_feedback(component, selection, template, feedback_set)
    mapping = component.substitute(item["enabled-set-feedback-mapping"])
    for key in selection.enabled_set:
        if key in mapping:
            feedback_set.add(mapping[key])


def _prepare_components(director: PackageBuildDirector,
                        component: WorkspaceComponent,
                        upper_selection: ItemSelection) -> EnabledSet:
    # Add the enabled set of the upper level component to our own enabled set
    selection = component.selection
    logging.info("%s: initial enabled set: %s", component.uid,
                 sorted(selection.enabled_set))
    selection.extend_enabled_set(upper_selection.enabled_set)
    logging.info("%s: enabled set after upper level extension: %s",
                 component.uid, sorted(selection.enabled_set))

    # Apply enabled set actions of templates and us
    _apply_enabled_set_actions(component, selection, component.item)

    # Add enabled set feedback from templates and us
    feedback_set: set[str] = set()
    _gather_feedback(component, selection, component.item, feedback_set)
    selection.extend_enabled_set(feedback_set)
    logging.info("%s: enabled set after template and own feedback: %s",
                 component.uid, sorted(selection.enabled_set))

    # Prepare next level components and get their enabled set feedback
    for link in component.item.links_to_children("input"):
        if link.item.get("workspace-type", None) == "component":
            subcomponent = director[link.item.uid]
            assert isinstance(subcomponent, WorkspaceComponent)
            feedback_set.update(
                _prepare_components(director, subcomponent, selection))
    selection.extend_enabled_set(feedback_set)
    logging.info("%s: enabled set after lower level feedback: %s",
                 component.uid, sorted(selection.enabled_set))

    component.item["enabled-set"] = sorted(selection.enabled_set)
    logging.info("%s: enabled set feedback for upper level: %s", component.uid,
                 sorted(feedback_set))
    return feedback_set


def _expand_templates(director: PackageBuildDirector,
                      component: WorkspaceComponent) -> None:
    with component.selection_and_view_scope():
        new_uids: list[str] = []
        for item in itertools.chain(
                component.item.parents(("use-package-template",
                                        "use-package-component-template")),
                component.item.children(("add-package-template",
                                         "add-package-component-template"))):
            template = director[item.uid]
            template.expand_template(component, new_uids)
        director.item_cache.initialize_links(new_uids)
        for link in component.item.links_to_children("input"):
            if link.item.get("workspace-type", None) == "component":
                subcomponent = director[link.item.uid]
                assert isinstance(subcomponent, WorkspaceComponent)
                _expand_templates(director, subcomponent)


@dataclasses.dataclass
class WorkspaceConfig:
    """ Represents a workspace configuration. """
    # pylint: disable=too-many-instance-attributes
    spec_directories: list[str]
    workspace_directory: str = "."
    package_uid: str = "/pkg/component"
    cache_directory: str = "cache-workspace"
    enabled_set: list[str] = dataclasses.field(default_factory=list)
    verify_specification_format: bool = True
    factory: BuildItemFactory | None = None
    extra_type_data_by_uid: ItemDataByUID = dataclasses.field(
        default_factory=dict)


@dataclasses.dataclass
class WorkspaceContext:
    """
    Provides a context with a workspace director and associated elements.
    """
    cache: ItemCache
    director: PackageBuildDirector
    workspace_directory: str
    extra_type_data_by_uid: ItemDataByUID


def create_workspace(config: WorkspaceConfig) -> WorkspaceContext:
    """ Create a workspace package build director. """
    cache_config = ItemCacheConfig(enabled_set=config.enabled_set,
                                   paths=config.spec_directories,
                                   resolve_proxies=False,
                                   cache_directory=config.cache_directory)
    type_provider = BuildItemTypeProvider(config.extra_type_data_by_uid)
    item_cache = ItemCache(cache_config, type_provider=type_provider)
    factory = config.factory
    if factory is None:
        factory = create_workspace_item_factory()
    director = PackageBuildDirector(item_cache, config.package_uid, factory)
    package = director.package
    package.item.view["workspace-commit"] = _git_commit(
        config.workspace_directory)
    package.item.view["workspace-directory"] = config.workspace_directory
    assert isinstance(package, WorkspaceComponent)
    _prepare_components(director, package, item_cache.active_selection)
    _expand_templates(director, package)
    _verify_specification_format(item_cache,
                                 config.verify_specification_format)
    return WorkspaceContext(item_cache, director, config.workspace_directory,
                            config.extra_type_data_by_uid)


@dataclasses.dataclass
class _UIDs:
    # pylint: disable=too-many-instance-attributes
    deleted: list[str]
    workspace: list[str]
    workspace_modified: list[str]
    workspace_copy: list[str]
    workspace_dummy: list[str]
    other_existing: list[str]
    other_new: list[str]
    other_modified: list[str]

    def __init__(self, workspace_cache: ItemCache,
                 buildspace_cache: ItemCache) -> None:
        spec_types: set[str] = set(
            item.uid for item in workspace_cache.items_by_type["spec"])
        workspace_all: set[str] = set(
            itertools.chain(workspace_cache.keys(),
                            workspace_cache.proxies.keys())) - spec_types
        buildspace_all: set[str] = set(
            itertools.chain(buildspace_cache.keys(),
                            buildspace_cache.proxies.keys())) - spec_types

        # Get UIDs of all workspace type items
        workspace: set[str] = set()
        for name in workspace_cache.types:
            if name.startswith("pkg/workspace/"):
                workspace.update(
                    item.uid for item in workspace_cache.items_by_type[name])

        # Get the dependencies of all workspace type items to sort them
        # topologically
        workspace_deps: dict[str, set[str]] = {}
        for uid in workspace:
            workspace_deps.setdefault(uid, set())
            item = workspace_cache[uid]
            for parent in item.parents("input",
                                       is_link_enabled=link_is_enabled):
                if parent.uid in workspace:
                    workspace_deps[uid].add(parent.uid)
            for parent in item.parents("input-to",
                                       is_link_enabled=link_is_enabled):
                if parent.uid in workspace:
                    workspace_deps.setdefault(parent.uid, set()).add(uid)
        workspace_deps = dict(sorted(workspace_deps.items()))

        self.deleted = sorted(buildspace_all - workspace_all)
        self.workspace = list(
            graphlib.TopologicalSorter(workspace_deps).static_order())
        self.workspace_modified = []
        self.workspace_copy = []
        self.workspace_dummy = []
        other = workspace_all - workspace
        self.other_existing = sorted(other & buildspace_all)
        self.other_new = sorted(other - buildspace_all)
        self.other_modified = []
        assert set(
            itertools.chain(self.other_new, self.other_existing,
                            self.workspace)) == workspace_all


@dataclasses.dataclass
class BuildspaceConfig:
    """ Represents a buildspace configuration. """
    spec_directory: str
    cache_directory: str = "cache-buildspace"
    use_git: bool = False
    verify_specification_format: bool = False
    factory: BuildItemFactory | None = None


@dataclasses.dataclass
class BuildspaceContext:
    """
    Provides a context with a buildspace director and associated elements.
    """
    cache: ItemCache
    director: PackageBuildDirector
    spec_directory: str


def _create_buildspace(
        config: BuildspaceConfig, package_uid: str, enabled_set: list[str],
        extra_type_data_by_uid: ItemDataByUID) -> BuildspaceContext:
    os.makedirs(config.spec_directory, exist_ok=True)
    cache_config = ItemCacheConfig(enabled_set=enabled_set,
                                   paths=[config.spec_directory],
                                   resolve_proxies=False,
                                   cache_directory=config.cache_directory)
    type_provider = BuildItemTypeProvider(extra_type_data_by_uid)
    item_cache = JSONItemCache(cache_config, type_provider=type_provider)
    factory = config.factory
    if factory is None:
        factory = create_build_item_factory()
    director = PackageBuildDirector(item_cache, package_uid, factory,
                                    config.use_git)
    return BuildspaceContext(item_cache, director, config.spec_directory)


def _gather_workspace_modified(workspace_director: PackageBuildDirector,
                               buildspace_cache: ItemCache,
                               uids: _UIDs) -> None:
    for uid in uids.workspace:
        workspace_item = workspace_director[uid]
        assert isinstance(workspace_item, _WorkspaceItem)
        with workspace_item.selection_and_view_scope():
            if workspace_item.item.enabled:
                logging.info("%s: load workspace state", uid)
                digest = workspace_item.load_workspace_state()
                is_dummy = False
            else:
                logging.info("%s: use dummy for disabled item", uid)
                dummy_digest = _DUMMY_DATA["workspace-hash"]
                assert isinstance(dummy_digest, str)
                digest = dummy_digest
                is_dummy = True
        workspace_item["workspace-hash"] = digest
        buildspace_item = buildspace_cache.get(uid, None)
        if buildspace_item is not None and digest == buildspace_item[
                "workspace-hash"]:
            continue
        if buildspace_item is not None:
            uids.workspace_modified.append(uid)
        if is_dummy:
            uids.workspace_dummy.append(uid)
        else:
            uids.workspace_copy.append(uid)
            data = workspace_item.get_buildspace_data()
            data["workspace-hash"] = workspace_item["workspace-hash"]
            workspace_item.item.view["buildspace-data"] = data


def _gather_other_modified(workspace_director: PackageBuildDirector,
                           buildspace_cache: ItemCache, uids: _UIDs) -> None:
    for uid in uids.other_existing:
        workspace_item = workspace_director.item_cache[uid]
        digest = _workspace_hash(workspace_item.data)
        if digest != buildspace_cache[uid]["workspace-hash"]:
            workspace_item["workspace-hash"] = digest
            uids.other_modified.append(uid)


def _set_other_workspace_hash(workspace_cache: ItemCache,
                              other_new: list[str]) -> None:
    for uid in other_new:
        workspace_item = workspace_cache[uid]
        workspace_item["workspace-hash"] = _workspace_hash(workspace_item.data)


def _delete_buildspace_items(director: PackageBuildDirector,
                             deleted_uids: list[str],
                             item_files: list[str]) -> None:
    for uid in deleted_uids:
        buildspace_item = director[uid]
        logging.info("%s: discard buildspace state", uid)
        buildspace_item.discard()
        buildspace_item.clear()
        buildspace_item.commit("Delete")
        file = buildspace_item.item.file
        director.remove(uid)
        logging.info("%s: remove: %s", uid, file)
        os.remove(file)
        item_files.append(file)


def _add_buildspace_submodules(
        workspace_package: PackageComponent,
        buildspace_director: PackageBuildDirector) -> None:
    for uid in workspace_package.view.get("submodules", []):
        try:
            # Creating a repository state will register it as a submodule
            buildspace_director[uid]
        except KeyError:
            pass


def _discard_buildspace_items(director: PackageBuildDirector,
                              discard_uids: list[str]) -> None:
    for uid in discard_uids:
        buildspace_item = director[uid]
        logging.info("%s: discard buildspace state", uid)
        buildspace_item.discard()
        buildspace_item.clear()
        buildspace_item.commit("Discard")
        director.remove(uid)


def _add_buildspace_item(buildspace: BuildspaceContext, uid: str,
                         data: dict) -> Item:
    new_item = buildspace.cache.add_item(uid, data, initialize_links=False)
    file = os.path.join(buildspace.spec_directory, f"{uid[1:]}.json")
    new_item.file = file
    os.makedirs(os.path.dirname(file), exist_ok=True)
    return new_item


def _copy_to_buildspace(workspace_director: PackageBuildDirector,
                        buildspace_director: PackageBuildDirector,
                        workspace_copy: list[str]) -> None:
    for uid in workspace_copy:
        workspace_item = workspace_director[uid]
        assert isinstance(workspace_item, _WorkspaceItem)
        with workspace_item.selection_and_view_scope():
            buildspace_item = buildspace_director[uid]
            logging.info("%s: copy to buildspace", uid)
            workspace_item.copy_to_buildspace(buildspace_item)
            buildspace_item.commit("Copy to buildspace")


def export_to_buildspace(workspace: WorkspaceContext,
                         config: BuildspaceConfig) -> BuildspaceContext:
    """ Export the workspace to the buildspace. """
    buildspace = _create_buildspace(config, workspace.director.package_uid,
                                    list(workspace.cache.enabled_set),
                                    workspace.extra_type_data_by_uid)
    uids = _UIDs(workspace.cache, buildspace.cache)
    _gather_workspace_modified(workspace.director, buildspace.cache, uids)
    _gather_other_modified(workspace.director, buildspace.cache, uids)
    _set_other_workspace_hash(workspace.cache, uids.other_new)
    item_files: list[str] = []
    _delete_buildspace_items(buildspace.director, uids.deleted, item_files)
    _add_buildspace_submodules(workspace.director.package, buildspace.director)

    # Discard other items first and the modified workspace items in reverse
    # order since we may have modified the package component which provides the
    # Git directory
    _discard_buildspace_items(buildspace.director, uids.other_modified)
    _discard_buildspace_items(buildspace.director,
                              sorted(uids.workspace_modified, reverse=True))

    for uid in uids.workspace_copy:
        _add_buildspace_item(buildspace, uid,
                             workspace.cache[uid].view["buildspace-data"])
    for uid in uids.workspace_dummy:
        item = _add_buildspace_item(buildspace, uid, _DUMMY_DATA)
        item.save()
        item_files.append(item.file)
    for uid in itertools.chain(uids.other_new, uids.other_modified):
        item = _add_buildspace_item(buildspace, uid,
                                    copy.deepcopy(workspace.cache[uid].data))
        item.save()
        item_files.append(item.file)
    buildspace.cache.reinitialize_links()
    _copy_to_buildspace(workspace.director, buildspace.director,
                        uids.workspace_copy)
    if item_files:
        package = buildspace.director.package
        package.git_add(item_files)
        package.commit("Update buildspace items")
    buildspace.cache.resolve_proxies = True
    _verify_specification_format(buildspace.cache,
                                 config.verify_specification_format)
    return buildspace



# SPDX-License-Identifier: BSD-2-Clause
""" Provides a command line interface to update a performance limits item. """

# Copyright (C) 2023, 2026 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import argparse
import json
import sys

from specitems import (Item, ItemCache, ItemCacheConfig,
                       get_item_cache_arguments)

from .testoutputparser import augment_report

_LimitsByUID = dict[str, dict[str, dict[str, float]]]


def _calculate_limits(runtime_measurement: dict[str, float], uid: str,
                      variant: str, limits_by_uid: _LimitsByUID,
                      args: argparse.Namespace) -> dict[str, float]:
    minimum = runtime_measurement["min"]
    median = runtime_measurement["q2"]
    maximum = runtime_measurement["max"]
    min_lower_bound = args.lower_bound_scaler * minimum
    max_upper_bound = args.upper_bound_scaler * maximum
    median_lower_bound = max(args.lower_bound_scaler * median, min_lower_bound)
    median_upper_bound = min(args.upper_bound_scaler * median, max_upper_bound)
    if args.lazy:
        try:
            current_limits = limits_by_uid[uid][variant]
        except KeyError:
            pass
        else:
            limit = current_limits["min-lower-bound"]
            if limit <= minimum:
                min_lower_bound = limit
            limit = current_limits["median-lower-bound"]
            if limit <= median:
                median_lower_bound = limit
            limit = current_limits["median-upper-bound"]
            if limit >= median:
                median_upper_bound = limit
            limit = current_limits["max-upper-bound"]
            if limit >= maximum:
                max_upper_bound = limit
    return {
        "max-upper-bound": max_upper_bound,
        "median-lower-bound": median_lower_bound,
        "median-upper-bound": median_upper_bound,
        "min-lower-bound": min_lower_bound
    }


def _update_perf_limits(reports: list[dict], name_to_uid: dict[str, str],
                        limits_by_uid: _LimitsByUID,
                        args: argparse.Namespace) -> None:
    for report in reports:
        if "performance" in report["executable"]:
            if "info" not in report:
                augment_report(report, report["output"])
            try:
                test_cases = report["test-suite"]["test-cases"]
            except KeyError:
                print("no performance data available for:",
                      report["executable"])
                continue
            for test_case in test_cases:
                for runtime_measurement in test_case["runtime-measurements"]:
                    uid = name_to_uid[runtime_measurement["name"]]
                    variant = runtime_measurement["variant"]
                    limits_by_uid.setdefault(uid,
                                             {})[variant] = _calculate_limits(
                                                 runtime_measurement, uid,
                                                 variant, limits_by_uid, args)


def _write_perf_limits(perf_limits: Item, limits_by_uid: _LimitsByUID) -> None:
    links = [
        link for link in perf_limits["links"]
        if link["role"] != "performance-runtime-limits"
    ]
    for uid, limits in sorted(limits_by_uid.items()):
        links.append({
            "limits": limits,
            "role": "performance-runtime-limits",
            "uid": uid
        })
    perf_limits.data["links"] = links
    perf_limits.save()


def cliupdateperf(argv: list[str] = sys.argv) -> None:
    """ Update a performance limits item using test results. """

    def _add_arguments(parser):
        parser.add_argument("--lazy",
                            action="store_true",
                            help="only update necessary limits")
        parser.add_argument(
            "--lower-bound-scaler",
            type=float,
            default=0.9,
            help="scaler to define lower bounds (default: 0.9)")
        parser.add_argument(
            "--upper-bound-scaler",
            type=float,
            default=1.1,
            help="scaler to define upper bounds (default: 1.1)")
        parser.add_argument(
            "perf_limits_uid",
            metavar="PERF_LIMITS_UID",
            help="the UID of a performance limits item to update")
        parser.add_argument("reports",
                            metavar="REPORT",
                            nargs="+",
                            help="a test report JSON file")

    args = get_item_cache_arguments(argv[1:],
                                    description=cliupdateperf.__doc__,
                                    add_arguments=(_add_arguments, ))
    cache_config = ItemCacheConfig(paths=args.spec_directories,
                                   cache_directory=args.cache_directory,
                                   initialize_links=False,
                                   resolve_proxies=False)
    item_cache = ItemCache(cache_config)
    name_to_uid: dict[str, str] = {}
    for item in item_cache.values():
        if item.get("type") != "requirement":
            continue
        if item["requirement-type"] != "non-functional":
            continue
        if item["non-functional-type"] == "performance-runtime":
            name_to_uid[item.ident] = item.uid
    perf_limits = item_cache[args.perf_limits_uid]
    limits_by_uid: _LimitsByUID = {}
    for link in perf_limits["links"]:
        if link["role"] == "performance-runtime-limits":
            limits_by_uid[link["uid"]] = link["limits"]
    for report in args.reports:
        with open(report, "r", encoding="utf-8") as src:
            data = json.load(src)
        _update_perf_limits(data["reports"], name_to_uid, limits_by_uid, args)
    _write_perf_limits(perf_limits, limits_by_uid)



# SPDX-License-Identifier: BSD-2-Clause
""" Provides package change descriptions. """

# Copyright (C) 2023, 2026 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import itertools

from specitems import Item, ItemMapper, SphinxContent

from .pkgitems import BuildItem


def _issue_prologue(status: str) -> str:
    return ("At the time when the package of this version was produced, "
            f"the following {status} were present.")


def _add_issues(content: SphinxContent, mapper: ItemMapper, issues: set[Item],
                header: str, prologue: str, which: str) -> None:
    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-positional-arguments
    with content.section(header):
        if issues:
            content.add(prologue)
            rows: list[tuple[str,
                             ...]] = [("Database", "Identifier", "Subject")]
            for item in sorted(issues):
                database = item.parent("issue-member")
                url = mapper.substitute(database["url"], item=item)
                identifier = f"`{item['identifier']} <{url}>`_"
                subject = item["subject"].replace("`", "\\`")
                rows.append((database["name"], identifier, subject))
            content.add_grid_table(rows, widths=[27, 14, 59], font_size=-2)
        else:
            content.add(f"""At the time when the package was produced,
there were no {which} associated.""")


class PackageChanges(BuildItem):
    """ Provides package change descriptions. """

    def _get_issues(self, change: dict[str,
                                       str]) -> tuple[set[Item], set[Item]]:
        issues: tuple[set[Item], set[Item]] = (set(), set())
        package_status = self.item.map(change["package-status"])
        for link in package_status.links_to_parents("issue"):
            status = link["status"]
            if status != "N/A":
                issues[int(status == "open")].add(link.item)
        return issues

    def _get_change_list(self, mapper: ItemMapper, section_level: int,
                         with_description: bool) -> list[SphinxContent]:
        # pylint: disable=too-many-locals
        change_list: list[SphinxContent] = []
        past_issues: tuple[set[Item], set[Item]] = (set(), set())
        previous_issues: tuple[set[Item], set[Item]] = (set(), set())
        for index, change in enumerate(
                mapper.substitute_data(self.item["change-list"],
                                       item=self.item)):
            for past, previous in zip(past_issues, previous_issues):
                past.update(previous)
            content = SphinxContent(section_level=section_level)
            if with_description:
                content.add_blank_line()
                content.open_section(change["name"])
                content.add(change["description"])
            current_issues = self._get_issues(change)
            new_issues = current_issues[1].difference(previous_issues[1])
            open_issues = current_issues[1].intersection(previous_issues[1])
            closed_issues = current_issues[0].difference(past_issues[0])
            if index == 0:
                assert not open_issues
                _add_issues(
                    content, mapper, new_issues, "Initially open issues",
                    "For the initial package version, "
                    "the following issues were open.", "open")
            else:
                status = ("newly open issues with respect to "
                          "the previous package version")
                _add_issues(content, mapper, new_issues, "Newly open issues",
                            _issue_prologue(status), status)
                status = ("open issues which were newly or already open "
                          "in the previous package version")
                _add_issues(content, mapper,
                            open_issues, "Already open issues",
                            _issue_prologue(status), status)
                _add_issues(
                    content, mapper, closed_issues, "Closed issues",
                    "For this package version, the following "
                    "issues which were newly or already open in "
                    "the previous package version were closed.", "closed")
            if with_description:
                content.close_section()
            change_list.insert(0, content)
            previous_issues = current_issues
        return change_list

    def get_change_list(self, mapper: ItemMapper, section_level: int) -> str:
        """ Get the change list using the section level. """
        change_list = self._get_change_list(mapper, section_level, True)
        return "\n".join(itertools.chain.from_iterable(change_list))

    def get_open_issues(self, mapper: ItemMapper, section_level: int) -> str:
        """ Get the open issues using the section level. """
        content = SphinxContent(section_level=section_level)
        _, open_issues = self._get_issues(
            mapper.substitute_data(self.item["change-list"][-1],
                                   item=self.item))
        _add_issues(content, mapper, open_issues, "Open issues",
                    _issue_prologue("open issues"), "open")
        return "\n".join(content)

    def get_current_changes(self, mapper: ItemMapper) -> str:
        """ Get the current changes. """
        return mapper.substitute(self.item["change-list"][-1]["description"],
                                 item=self.item)

    def get_current_issues(self, mapper: ItemMapper,
                           section_level: int) -> str:
        """ Get the current issues using the section level. """
        change_list = self._get_change_list(mapper, section_level, False)
        return "\n".join(change_list[0])



# SPDX-License-Identifier: BSD-2-Clause
""" Generates images for special test programs of the RTEMS test suites. """

# Copyright (C) 2024, 2025 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import json
import logging
import math
import os
import re
import statistics
from typing import Any, Callable
import matplotlib.pyplot as plt  # type: ignore
from matplotlib import ticker  # type: ignore

from specitems import Item

from .directorystate import DirectoryState
from .pkgitems import PackageBuildDirector
from .testaggregator import TestAggregator

logging.getLogger("matplotlib").setLevel(logging.WARNING)


def _default_files(base: str) -> list[str]:
    return [base]


def _savefig(fig: plt.Figure, axes: plt.Axes, base: str) -> list[str]:
    axes.legend(loc="best")
    fig.tight_layout()
    plt.savefig(f"{base}.png")
    plt.savefig(f"{base}.pdf")
    return [f"{base}.png", f"{base}.pdf"]


def _tmfine01_build(title: str, base: str, data: dict) -> list[str]:
    fig, axes = plt.subplots()
    axes.set_title(f"{title}\nUncontested Mutex Performance")
    axes.set_xlabel("Active Workers")
    axes.set_ylabel("Operation Count")
    x = list(range(1, len(data[0]["counter"]) + 1))
    axes.xaxis.set_major_locator(ticker.FixedLocator(x))
    for samples in data:
        if samples["type"] != "private-mutex":
            continue
        y = [sum(values) for values in samples["counter"]]
        axes.plot(x,
                  y,
                  label=samples["description"].replace(
                      "Obtain/Release Private ", ""),
                  marker="o")
    return _savefig(fig, axes, base)


def _tmcontext01_build(title: str, base: str, data: dict) -> list[str]:
    fig, axes = plt.subplots()
    axes.set_title(f"{title}\nContext Switch Timing Test")
    axes.set_xlabel("Function Nest Level")
    axes.set_ylabel("Context Switch Time [μs]")
    x = list(range(0, len(data[0]["stats-by-function-nest-level"])))
    axes.xaxis.set_major_locator(ticker.FixedLocator(x))
    for samples in data:
        y = [
            values[2] / 1000.0
            for values in samples["stats-by-function-nest-level"]
        ]
        axes.plot(x, y, label=samples["environment"], marker="o")
    return _savefig(fig, axes, base)


def _tmtimer01_build(title: str, base: str, data: dict) -> list[str]:
    fig, axes = plt.subplots()
    axes.set_title(f"{title}\nTimer Fire and Cancel Timing Test")
    axes.set_xlabel("Active Timers")
    axes.set_xscale("log")
    axes.set_ylabel("Timer Fire and Cancel Duration [μs]")
    x = [sample["active-timers"] for sample in data["samples"]]
    for key in ["first", "middle", "last"]:
        y = [sample[key] / 1000.0 for sample in data["samples"]]
        axes.plot(x, y, label=f"operate on {key} timer", marker="o")
    return _savefig(fig, axes, base)


def _smpopenmp01_build(title: str, base: str, data: dict) -> list[str]:
    fig, axes = plt.subplots()
    axes.set_title(f"{title}\nOpenMP Microbench")
    axes.set_xlabel("Number of Threads")
    axes.set_ylabel("Relative Duration")
    x = list(range(1, len(data) + 1))
    axes.xaxis.set_major_locator(ticker.FixedLocator(x))
    for key in [
            "barrier-bench", "dynamic-bench", "guided-bench", "parallel-bench",
            "runtime-bench", "single-bench", "static-bench"
    ]:
        d = [results[key] for results in data]
        y = [x / d[0] for x in d]
        axes.plot(x, y, label=key.replace("-bench", ""), marker="o")
    return _savefig(fig, axes, base)


def _normed_coefficient_of_variation(counter: list[int]) -> float:
    return (statistics.stdev(counter) / statistics.mean(counter)) / math.sqrt(
        len(counter))


def _smplock01_files(base: str) -> list[str]:
    return [f"{base}-fair", f"{base}-perf"]


def _smplock01_fair(title: str, base: str, data: dict) -> list[str]:
    fig, axes = plt.subplots()
    axes.set_title(f"{title}\nSMP Lock Fairness")
    axes.set_xlabel("Active Workers")
    axes.set_ylabel("Normed Coefficient of Variation")
    axes.set_yscale("symlog", linthresh=1e-6)
    x = list(range(2, len(data[0]["results"]) + 1))
    axes.xaxis.set_major_locator(ticker.FixedLocator(x))
    for samples in data:
        if samples["lock-object"] != "global":
            continue
        if samples["section-type"] != "local counter":
            continue
        y = [
            _normed_coefficient_of_variation(results["counter"])
            for results in samples["results"][1:]
        ]
        axes.plot(x, y, label=samples["lock-type"], marker="o")
    return _savefig(fig, axes, f"{base}-fair")


def _smplock01_perf(title: str, base: str, data: dict) -> list[str]:
    fig, axes = plt.subplots()
    axes.set_title(f"{title}\nSMP Lock Performance")
    axes.set_xlabel("Active Workers")
    axes.set_ylabel("Operation Count")
    x = list(range(1, len(data[0]["results"]) + 1))
    axes.xaxis.set_major_locator(ticker.FixedLocator(x))
    for samples in data:
        if samples["lock-object"] != "global":
            continue
        if samples["section-type"] != "local counter":
            continue
        y = [sum(results["counter"]) for results in samples["results"]]
        axes.plot(x, y, label=samples["lock-type"], marker="o")
    return _savefig(fig, axes, f"{base}-perf")


def _smplock01_build(title: str, base: str, data: dict) -> list[str]:
    files = _smplock01_fair(title, base, data)
    files.extend(_smplock01_perf(title, base, data))
    return files


def _sptimecounter02_build(title: str, base: str, data: dict) -> list[str]:
    fig, axes = plt.subplots()
    axes.set_title(f"{title}\nTimestamp Performance")
    axes.set_xlabel("Active Workers")
    axes.set_ylabel("Operation Count")
    x = list(range(1, len(data[0]["counter"]) + 1))
    axes.xaxis.set_major_locator(ticker.FixedLocator(x))
    for samples in data:
        y = [sum(values) for values in samples["counter"]]
        axes.plot(x, y, label=samples["timecounter"], marker="o")
    return _savefig(fig, axes, base)


_IMAGE_GENERATORS = {
    "/build/testsuites/tmtests/tmfine01": (_default_files, _tmfine01_build),
    "/build/testsuites/tmtests/tmcontext01":
    (_default_files, _tmcontext01_build),
    "/build/testsuites/tmtests/tmtimer01": (_default_files, _tmtimer01_build),
    "/build/testsuites/smptests/smpopenmp01":
    (_default_files, _smpopenmp01_build),
    "/build/testsuites/smptests/smplock01":
    (_smplock01_files, _smplock01_build),
    "/build/testsuites/sptests/sptimecounter02":
    (_default_files, _sptimecounter02_build)
}


def _append_test_images(_title: str, base: str, report: dict,
                        generate: tuple) -> None:
    if _get_json_data(report) is not None:
        report.setdefault("images", []).extend(generate[0](base))


_JSON_DATA = re.compile(r"\*\*\* BEGIN OF JSON DATA \*\*\*(.*)"
                        r"\*\*\* END OF JSON DATA \*\*\*")


def _get_json_data(report: dict) -> Any:
    match = _JSON_DATA.search("".join(report["output"]))
    if match:
        return match.group(1)
    return None


class BuildRTEMSTestsImages(DirectoryState):
    """
    Generates images for special test programs of the RTEMS test suites.
    """

    def __init__(self, director: PackageBuildDirector, item: Item):
        super().__init__(director, item)
        self._generate(_append_test_images)

    def run(self) -> None:
        self.discard()
        self.clear()
        os.makedirs(self.directory, exist_ok=True)
        self._generate(self._build_test_images)
        plt.close()
        self.description.add(f"""Produce RTEMS tests images in directory
{self.description.path(self.directory)}""")

    def _build_test_images(self, title: str, base: str, report: dict,
                           generate: tuple) -> None:
        json_data = _get_json_data(report)
        if json_data is not None:
            logging.info("%s: generate: %s", self.uid, base)
            self.add_files(generate[1](title, base, json.loads(json_data)))
        else:
            logging.warning("%s: no JSON data for: %s", self.uid, base)

    def _generate(self, work: Callable[[str, str, dict, tuple], None]) -> None:
        tests = self.input("test-aggregation")
        assert isinstance(tests, TestAggregator)
        for target_data in tests.targets.values():
            target = target_data["key"]
            target_name = target_data["name"]
            for config_data in target_data["configs"]:
                config_key = config_data["config-key"]
                for uid, generate in _IMAGE_GENERATORS.items():
                    try:
                        report = config_data["test-programs"][uid]
                    except KeyError:
                        continue
                    name = f"{target}-{config_key}-{uid[1:]}".replace("/", "-")
                    base = os.path.join(self.directory, name)
                    title = f"""spec:{uid}
Target - {target_name}
Configuration - {config_key}"""
                    work(title, base, report, generate)



# SPDX-License-Identifier: BSD-2-Clause
""" Provides package build template items. """

# Copyright (C) 2025, 2026 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import itertools
import logging
import os
import re
from typing import Any, NamedTuple

from specitems import (IS_ENABLED_OPS, Item, is_enabled_with_ops, items,
                       to_iterable)

from . import pkgitems


class _AttributeActionContext(NamedTuple):
    item: pkgitems.BuildItem
    action: dict
    action_value: Any
    value: Any
    name: str
    idx: int
    path: str


def _attribute_action_append(ctx: _AttributeActionContext) -> None:
    logging.info("%s: at %s append: %s", ctx.item.uid, ctx.path,
                 ctx.action_value)
    if ctx.idx >= 0:
        ctx.value[ctx.name][ctx.idx].append(ctx.action_value)
    else:
        ctx.value[ctx.name].append(ctx.action_value)


def _attribute_action_extend(ctx: _AttributeActionContext) -> None:
    logging.info("%s: at %s extend by: %s", ctx.item.uid, ctx.path,
                 ctx.action_value)
    if ctx.idx >= 0:
        ctx.value[ctx.name][ctx.idx] += ctx.action_value
    else:
        ctx.value[ctx.name] += ctx.action_value


def _attribute_action_insert(ctx: _AttributeActionContext,
                             adjustment: int) -> None:
    if ctx.idx >= 0:
        the_list = ctx.value[ctx.name][ctx.idx]
    else:
        the_list = ctx.value[ctx.name]
    for index, the_value in enumerate(the_list):
        if the_value[ctx.action["where-key"]] == ctx.action["where-value"]:
            index += adjustment
            logging.info(
                "%s: at %s insert at index "
                "%i where '%s' is '%s': %s", ctx.item.uid, ctx.path, index,
                ctx.action["where-key"], ctx.action["where-value"],
                ctx.action_value)
            the_list.insert(index, ctx.action_value)
            return
    logging.info("%s: at %s no matching insert position", ctx.item.uid,
                 ctx.path)


def _attribute_action_insert_after(ctx: _AttributeActionContext) -> None:
    _attribute_action_insert(ctx, 1)


def _attribute_action_insert_before(ctx: _AttributeActionContext) -> None:
    _attribute_action_insert(ctx, 0)


def _attribute_action_remove(ctx: _AttributeActionContext) -> None:
    msg = f"{ctx.item.uid}: at {ctx.action['path']} remove: "
    if ctx.idx >= 0:
        logging.info("%s%s", msg, ctx.value[ctx.name][ctx.idx])
        del ctx.value[ctx.name][ctx.idx]
    else:
        logging.info("%s%s", msg, ctx.value[ctx.name])
        del ctx.value[ctx.name]


def _remove_list_items(ctx: _AttributeActionContext,
                       the_list: list[str]) -> list[str]:
    new_list: list[str] = []
    pattern = re.compile(ctx.action["pattern"])
    for value in the_list:
        if pattern.search(value) is None:
            new_list.append(value)
        else:
            logging.info("%s: at %s remove: %s", ctx.item.uid, ctx.path, value)
    return new_list


def _attribute_action_remove_list_items(ctx: _AttributeActionContext) -> None:
    if ctx.idx >= 0:
        ctx.value[ctx.name][ctx.idx] = _remove_list_items(
            ctx, ctx.value[ctx.name][ctx.idx])
    else:
        ctx.value[ctx.name] = _remove_list_items(ctx, ctx.value[ctx.name])


def _attribute_action_set(ctx: _AttributeActionContext) -> None:
    logging.info("%s: at %s set to: %s", ctx.item.uid, ctx.path,
                 ctx.action_value)
    if ctx.idx >= 0:
        ctx.value[ctx.name][ctx.idx] = ctx.action_value
    else:
        ctx.value[ctx.name] = ctx.action_value


def _attribute_action_set_default(ctx: _AttributeActionContext) -> None:
    if ctx.name in ctx.value:
        logging.info("%s: at %s no need for a default value", ctx.item.uid,
                     ctx.path)
    else:
        _attribute_action_set(ctx)


def _attribute_action_substitute(ctx: _AttributeActionContext) -> None:
    if ctx.idx >= 0:
        new_value = ctx.item.substitute(ctx.value[ctx.name][ctx.idx])
        ctx.value[ctx.name][ctx.idx] = new_value
    else:
        new_value = ctx.item.substitute(ctx.value[ctx.name])
        ctx.value[ctx.name] = new_value
    logging.info("%s: at %s substitute to: %s", ctx.item.uid, ctx.path,
                 new_value)


_ATTRIBUTE_ACTIONS = {
    "append": _attribute_action_append,
    "extend": _attribute_action_extend,
    "insert-after": _attribute_action_insert_after,
    "insert-before": _attribute_action_insert_before,
    "remove": _attribute_action_remove,
    "remove-list-items": _attribute_action_remove_list_items,
    "set": _attribute_action_set,
    "set-default": _attribute_action_set_default,
    "substitute": _attribute_action_substitute,
}


def _get_name_and_index(key: str) -> tuple[str, int]:
    parts = key.split("[")
    try:
        return parts[0], int(parts[1].split("]")[0])
    except IndexError:
        return parts[0], -1


def _attribute_action(item: pkgitems.BuildItem, action: dict,
                      data: Any) -> None:
    for path in to_iterable(action["path"]):
        path_list = path.strip("/").split("/")
        value = data
        for key in path_list[:-1]:
            name, index = _get_name_and_index(key)
            if index >= 0:
                value = value[name][index]
            else:
                value = value[name]
        name, index = _get_name_and_index(path_list[-1])
        _ATTRIBUTE_ACTIONS[action["action"]](_AttributeActionContext(
            item, action, item.substitute(action.get("value", None)), value,
            name, index, path))


def _enabled_by_has_type(item: Item, enabled_by: dict, who: str) -> bool:
    value = enabled_by[f"{who}-type"]
    if value is None:
        return True
    if isinstance(value, str):
        return item.type == value
    return item.type in value


class _Template(pkgitems.BuildItem):

    def run_attribute_actions(self, component: pkgitems.PackageComponent,
                              data: Any) -> None:
        """ Run the attribute actions on the data. """

        def _has_sibling(_enabled_set: list[str], enabled_by: Any,
                         _ops: dict) -> bool:
            for parent in component.item.parents(enabled_by["parent-role"]):
                if _enabled_by_has_type(parent, enabled_by, "parent"):
                    for child in parent.children(enabled_by["sibling-role"]):
                        if child == component.item:
                            continue
                        if _enabled_by_has_type(child, enabled_by, "sibling"):
                            return True
            return False

        ops = IS_ENABLED_OPS | {"has-sibling": _has_sibling}
        for action in self.item["attribute-actions"]:
            if not is_enabled_with_ops(component.selection.enabled_set,
                                       action["enabled-by"], ops):
                continue
            _attribute_action(component, action, data)

    def add_item(self, component: pkgitems.PackageComponent, data: dict,
                 new_uids) -> None:
        """ Add an item with the data along the component. """
        uid = f"{os.path.dirname(component.uid)}/{os.path.basename(self.uid)}"
        logging.info("%s: expand template to: %s", self.uid, uid)
        data = component.substitute(data)
        self.run_attribute_actions(component, data)
        component.item.cache.add_item(uid, data, initialize_links=False)
        new_uids.append(uid)

    def expand_template(self, component: pkgitems.PackageComponent,
                        new_uids: list[str]) -> None:
        """ Expand the template for the component. """
        for item in itertools.chain(
                self.item.parents("use-package-template"),
                self.item.children("add-package-template")):
            template_item = self.director[item.uid]
            template_item.expand_template(component, new_uids)


class ComponentTemplate(_Template):
    """ Provides a template for components. """

    def expand_template(self, component: pkgitems.PackageComponent,
                        new_uids: list[str]) -> None:
        self.run_attribute_actions(component, component.item.data)
        super().expand_template(component, new_uids)


class FileItemTemplate(_Template):
    """
    Provides a template for items where the attributes are provided by a file
    referenced by the template item.
    """

    def expand_template(self, component: pkgitems.PackageComponent,
                        new_uids: list[str]) -> None:
        path = component.substitute(self.item["file"])
        data = items.load_data(path)
        self.add_item(component, data, new_uids)
        super().expand_template(component, new_uids)


class InlineItemTemplate(_Template):
    """
    Provides a template for items where the attributes are provided by the
    template item.
    """

    def expand_template(self, component: pkgitems.PackageComponent,
                        new_uids: list[str]) -> None:
        data = self.item["attributes"]
        data["SPDX-License-Identifier"] = self.item["SPDX-License-Identifier"]
        data["copyrights"] = self.item["copyrights"]
        self.add_item(component, data, new_uids)
        super().expand_template(component, new_uids)



# SPDX-License-Identifier: BSD-2-Clause
""" Directory state with run actions support. """

# Copyright (C) 2025, 2026 embedded brains GmbH & Co. KG
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from typing import Optional

from specitems import Item
from specware import run_command

from .dirstatebase import DirectoryStateBase
from .pkgitems import BuildItemMapper, PackageBuildDirector
from .runactions import RunActionsProvider


class DirectoryState(DirectoryStateBase):
    """ Produces a directory state by running actions. """

    def __init__(self,
                 director: PackageBuildDirector,
                 item: Item,
                 mapper: Optional[BuildItemMapper] = None) -> None:
        super().__init__(director, item, mapper)
        self.run_actions_provider = RunActionsProvider(self)

    def run_actions(self) -> None:
        """ Run the optional actions. """
        self.run_actions_provider.run(self.item.get("actions", {}))

    def run_post_actions(self) -> None:
        """ Run the optional post-actions. """
        self.run_actions_provider.run(self.item.get("post-actions", {}))

    def run(self) -> None:
        super().run()
        self.run_actions()


class RepositoryState(DirectoryState):
    """ Maintains a repository state. """

    def __init__(self, director: PackageBuildDirector, item: Item):
        super().__init__(director, item)
        director.add_submodule(self.directory)

    def lazy_verify(self) -> bool:
        stdout: list[str] = []
        status = run_command(["git", "rev-parse", "HEAD"], self.directory,
                             stdout)
        return status == 0 and self["commit"] == stdout[0].strip()
